# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import unittest
from unittest import mock

import horizon
from horizon import exceptions

from muranodashboard.images import tables
from muranodashboard.images import views


class TestMarkedImagesView(unittest.TestCase):

    def setUp(self):
        super(TestMarkedImagesView, self).setUp()

        mock_request = mock.Mock(horizon={'async_messages': []})
        self.images_view = views.MarkedImagesView(request=mock_request)
        self.images_view._prev = False
        self.images_view._more = False

        self.assertEqual(tables.MarkedImagesTable,
                         self.images_view.table_class)
        self.assertEqual('images/index.html', self.images_view.template_name)
        self.assertEqual('Marked Images', self.images_view.page_title)

        mock_horizon_utils = mock.patch.object(views, 'utils').start()
        mock_horizon_utils.get_page_size.return_value = 2
        self.addCleanup(mock.patch.stopall)

    def _get_mock_image(self, prefix):
        image_info = {}
        if prefix:
            image_info = {
                "title": "{0}_title".format(prefix),
                "type": "{0}_type".format(prefix)
            }
        mock_image = mock.Mock(**{'murano_image_info': json.dumps(image_info)})
        return mock_image

    def test_has_prev_data(self):
        self.assertFalse(self.images_view.has_prev_data(None))

    def test_has_more_data(self):
        self.assertFalse(self.images_view.has_more_data(None))

    @mock.patch.object(views, 'glance', autospec=True)
    def test_get_data(self, mock_glance):
        """Test that get_data works."""
        foo_mock_image = self._get_mock_image('foo')
        bar_mock_image = self._get_mock_image('bar')
        # Filtered out by forms.filter_murano_images.
        mock_image_to_filter = self._get_mock_image(None)

        mock_glance_client = mock.Mock()
        mock_glance_client.images.list.return_value = [
            foo_mock_image, bar_mock_image, mock_image_to_filter]
        mock_glance.glanceclient.return_value = mock_glance_client

        self.images_view.request.GET.get.return_value = 'foo_marker'
        result = self.images_view.get_data()

        expected_images = [bar_mock_image, foo_mock_image]
        expected_kwargs = {
            'filters': {},
            'marker': 'foo_marker',
            'sort_dir': 'asc'
        }

        self.assertEqual(expected_images, result)
        self.assertTrue(self.images_view.has_more_data(None))
        self.assertTrue(self.images_view.has_prev_data(None))
        mock_glance_client.images.list.assert_called_once_with(
            **expected_kwargs)
        self.images_view.request.GET.get.assert_called_once_with(
            tables.MarkedImagesTable._meta.prev_pagination_param, None)
        mock_glance.glanceclient.assert_called_once_with(
            self.images_view.request)

    @mock.patch.object(views, 'glance', autospec=True)
    def test_get_data_with_desc_sort_dir(self, mock_glance):
        """Test that sorting in descending order works."""
        foo_mock_image = self._get_mock_image('foo')
        bar_mock_image = self._get_mock_image('bar')

        mock_glance_client = mock.Mock()
        mock_glance_client.images.list.return_value = [
            foo_mock_image, bar_mock_image]
        mock_glance.glanceclient.return_value = mock_glance_client

        self.images_view.request.GET.get.return_value = None
        result = self.images_view.get_data()

        expected_images = [foo_mock_image, bar_mock_image]
        expected_kwargs = {
            'filters': {},
            'sort_dir': 'desc'
        }

        self.assertEqual(expected_images, result)
        self.assertFalse(self.images_view.has_more_data(None))
        self.assertFalse(self.images_view.has_prev_data(None))
        mock_glance_client.images.list.assert_called_once_with(
            **expected_kwargs)
        self.images_view.request.GET.get.assert_has_calls([
            mock.call(tables.MarkedImagesTable._meta.prev_pagination_param,
                      None),
            mock.call(tables.MarkedImagesTable._meta.pagination_param, None)
        ])
        mock_glance.glanceclient.assert_called_once_with(
            self.images_view.request)

    @mock.patch.object(views, 'glance', autospec=True)
    def test_get_data_with_more_results(self, mock_glance):
        """Test that extra results are not included in return value."""
        foo_mock_image = self._get_mock_image('foo')
        bar_mock_image = self._get_mock_image('bar')
        extra_mock_image = self._get_mock_image('baz')  # Extra result.
        # Filtered out by forms.filter_murano_images.
        mock_image_to_filter = self._get_mock_image(None)

        mock_glance_client = mock.Mock()
        mock_glance_client.images.list.return_value = [
            foo_mock_image, bar_mock_image, extra_mock_image,
            mock_image_to_filter]
        mock_glance.glanceclient.return_value = mock_glance_client

        self.images_view.request.GET.get.return_value = 'foo_marker'
        result = self.images_view.get_data()

        # Extra result not included, and result should be reversed.
        expected_images = [bar_mock_image, foo_mock_image]
        expected_kwargs = {
            'filters': {},
            'marker': 'foo_marker',
            'sort_dir': 'asc'
        }

        self.assertEqual(expected_images, result)
        self.assertTrue(self.images_view.has_more_data(None))
        self.assertTrue(self.images_view.has_prev_data(None))
        mock_glance_client.images.list.assert_called_once_with(
            **expected_kwargs)
        self.images_view.request.GET.get.assert_called_once_with(
            tables.MarkedImagesTable._meta.prev_pagination_param, None)
        mock_glance.glanceclient.assert_called_once_with(
            self.images_view.request)

    @mock.patch.object(views, 'reverse', autospec=True)
    @mock.patch.object(views, 'glance', autospec=True)
    @mock.patch.object(horizon.messages, 'horizon_message_already_queued')
    def test_get_data_except_glance_image_list_exception(self, mock_messages,
                                                         mock_glance,
                                                         mock_reverse):
        """Test that glance_v1_client.images.list exception is handled."""
        mock_messages.return_value = True
        mock_glance_client = mock.Mock()
        mock_glance_client.images.list.side_effect = Exception()
        mock_glance.glanceclient.return_value = mock_glance_client
        mock_reverse.return_value = 'foo_reverse_url'
        self.images_view.request.GET.get.return_value = None

        with self.assertRaises(exceptions.Http302) as cm:
            self.images_view.get_data()
        e = cm.exception
        self.assertEqual('foo_reverse_url', e.location)

        mock_glance.glanceclient.assert_called_once_with(
            self.images_view.request)
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:catalog:index')



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from django.conf import settings
from django import forms
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from horizon import exceptions
from horizon import forms as horizon_forms
from horizon import messages
from openstack_dashboard.api import glance
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def filter_murano_images(images, request=None):
    # filter images by project owner
    filter_project = getattr(settings, 'MURANO_IMAGE_FILTER_PROJECT_ID', None)
    if filter_project:
        project_ids = [filter_project]
        if request:
            project_ids.append(request.user.tenant_id)
        images = filter(
            lambda x: getattr(x, 'owner', None) in project_ids, list(images))
    # filter out the snapshot image type
    images = filter(
        lambda x: getattr(x, 'image_type', None) != 'snapshot', list(images))
    marked_images = []
    for image in images:
        # Additional properties, whose value is always a string data type, are
        # only included in the response if they have a value.
        metadata = getattr(image, 'murano_image_info', None)
        if metadata:
            try:
                metadata = json.loads(metadata)
            except ValueError:
                msg = _('Invalid metadata for image: {0}').format(image.id)
                LOG.warning(msg)
                if request:
                    exceptions.handle(request, msg)
                metadata = {}
            image.title = metadata.get('title', 'No Title')
            image.type = metadata.get('type', 'No Type')

            marked_images.append(image)
    return marked_images


class MarkImageForm(horizon_forms.SelfHandlingForm):
    _metadata = {
        'windows.2012': ' Windows Server 2012',
        'linux': 'Generic Linux',
        'cirros.demo': 'CirrOS for Murano Demo',
        'custom': "Custom type"
    }

    image = forms.ChoiceField(label=_('Image'))
    title = forms.CharField(max_length="255", label=_("Title"))
    type = forms.ChoiceField(
        label=_("Type"),
        choices=_metadata.items(),
        initial='custom',
        widget=forms.Select(attrs={
            'class': 'switchable',
            'data-slug': 'type'}))
    custom_type = forms.CharField(
        max_length="255",
        label=_("Custom Type"),
        widget=forms.TextInput(attrs={
            'class': 'switched',
            'data-switch-on': 'type',
            'data-type-custom': _('Custom Type')}),
        required=False)
    existing_titles = forms.CharField(widget=forms.HiddenInput)

    def __init__(self, request, *args, **kwargs):
        super(MarkImageForm, self).__init__(request, *args, **kwargs)

        images = []
        try:
            # https://bugs.launchpad.net/murano/+bug/1339261 - glance
            # client version change alters the API. Other tuple values
            # are _more and _prev (in recent glance client)
            images = glance.image_list_detailed(request)[0]
        except Exception:
            LOG.error('Failed to request image list from Glance')
            exceptions.handle(request, _('Unable to retrieve list of images'))

        # filter out the image format aki and ari
        images = filter(
            lambda x: x.container_format not in ('aki', 'ari'), images)

        # filter out the snapshot image type
        images = filter(
            lambda x: x.properties.get("image_type", '') != 'snapshot', images)

        self.fields['image'].choices = [(i.id, i.name) for i in images]
        self.fields['existing_titles'].initial = \
            [image.title for image in filter_murano_images(images)]

    def handle(self, request, data):
        LOG.debug('Marking image with specified metadata: {0}'.format(data))

        image_id = data['image']
        image_type = data['type'] if data['type'] != 'custom' else \
            data['custom_type']
        kwargs = {}
        kwargs['murano_image_info'] = json.dumps({
            'title': data['title'],
            'type': image_type
        })
        try:
            img = glance.image_update_properties(request, image_id, **kwargs)
            messages.success(request, _('Image successfully marked'))
            return img
        except Exception:
            exceptions.handle(request, _('Unable to mark image'),
                              redirect=reverse(
                                  'horizon:app-catalog:images:index'))

    def clean_title(self):
        cleaned_data = super(MarkImageForm, self).clean()
        title = cleaned_data.get('title')
        existing_titles = self.fields['existing_titles'].initial
        if title in existing_titles:
            raise forms.ValidationError(_('Specified title already in use.'
                                          ' Please choose another one.'))

        return title



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
from io import BytesIO
import json
import logging
import os
import testtools
import time
import urllib.parse as urlparse
import uuid

from glanceclient import client as gclient
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session
from keystoneclient.v3 import client as ks_client
from muranoclient.common import exceptions as muranoclient_exc
from muranoclient.glance import client as glare_client
import muranoclient.v1.client as mclient
from oslo_log import handlers
from oslo_log import log
from selenium.common import exceptions as exc
from selenium import webdriver
import selenium.webdriver.common.by as by
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support import ui

from muranodashboard.tests.functional.config import config as cfg
from muranodashboard.tests.functional import consts
from muranodashboard.tests.functional import utils

logger = log.getLogger(__name__).logger
logger.level = logging.DEBUG
logger.addHandler(handlers.ColorHandler())


class UITestCase(testtools.TestCase):
    @classmethod
    def setUpClass(cls):
        auth = v3.Password(user_domain_name='Default',
                           username=cfg.common.user,
                           password=cfg.common.password,
                           project_domain_name='Default',
                           project_name=cfg.common.tenant,
                           auth_url=cfg.common.keystone_url)
        session = ks_session.Session(auth=auth)
        cls.keystone_client = ks_client.Client(session=session)
        cls.auth_ref = auth.get_auth_ref(session)
        cls.service_catalog = cls.auth_ref.service_catalog
        if utils.glare_enabled():
            glare_endpoint = "http://127.0.0.1:9494"
            artifacts_client = glare_client.Client(
                endpoint=glare_endpoint, token=cls.auth_ref.auth_token,
                insecure=False, key_file=None, ca_file=None, cert_file=None,
                type_name="murano", type_version=1)
        else:
            artifacts_client = None
        cls.murano_client = mclient.Client(
            artifacts_client=artifacts_client,
            endpoint_override=cfg.common.murano_url,
            session=session)

        cls.url_prefix = urlparse.urlparse(cfg.common.horizon_url).path or ''
        if cls.url_prefix.endswith('/'):
            cls.url_prefix = cls.url_prefix[:-1]

    def setUp(self):
        super(UITestCase, self).setUp()

        # Enables zip files to be automatically saved to disk, without opening
        # a browser dialog.
        fp = webdriver.FirefoxProfile()
        fp.set_preference("browser.download.folderList", 2)
        fp.set_preference("browser.download.manager.showWhenStarting", False)
        fp.set_preference("browser.download.dir", os.getcwd())
        fp.set_preference("browser.helperApps.neverAsk.saveToDisk",
                          "application/octet-stream")

        self.driver = webdriver.Firefox(firefox_profile=fp)
        self.addCleanup(self.driver.quit)
        self.driver.maximize_window()
        self.driver.get(cfg.common.horizon_url + '/app-catalog/environments')
        self.driver.implicitly_wait(30)
        self.addOnException(self.take_screenshot)
        self.log_in()
        self.switch_to_project(cfg.common.tenant)

    def tearDown(self):
        super(UITestCase, self).tearDown()

        for env in self.murano_client.environments.list():
            self.remove_environment(env.id)

    def gen_random_resource_name(self, prefix=None, reduce_by=None):
        random_name = str(uuid.uuid4()).replace('-', '')[::reduce_by]
        if prefix:
            random_name = prefix + '_' + random_name
        return random_name

    def remove_environment(self, environment_id, timeout=180):
        self.murano_client.environments.delete(environment_id)

        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                self.murano_client.environments.get(environment_id)
                time.sleep(1)
            except Exception:
                # TODO(smurashov): bug/1378764 replace Exception to NotFound
                return
        raise Exception(
            'Environment {0} was not deleted in {1} seconds'.format(
                environment_id, timeout))

    @classmethod
    def create_user(cls, name, password=None, email=None, tenant_id=None):
        if tenant_id is None:
            projects = cls.keystone_client.projects.list()
            tenant_id = [project.id for project in projects
                         if project.name == cfg.common.tenant][0]
            cls.keystone_client.users.create(name, domain='default',
                                             password=password,
                                             email=email,
                                             project=tenant_id,
                                             enabled=True)
        else:
            cls.keystone_client.users.create(name, domain='default',
                                             password=password,
                                             email=email,
                                             project=tenant_id,
                                             enabled=True)
        roles = cls.keystone_client.roles.list()
        role_id = [role.id for role in roles if role.name == 'member'][0]
        users = cls.keystone_client.users.list()
        user_id = [user.id for user in users if user.name == name][0]
        cls.keystone_client.roles.grant(role_id, user=user_id,
                                        project=tenant_id)

    @classmethod
    def delete_user(cls, name):
        cls.keystone_client.users.find(name=name).delete()

    def get_tenantid_by_name(self, name):
        """Returns TenantID of the project by project's name"""
        tenant_id = [tenant.id for tenant
                     in self.keystone_client.projects.list()
                     if tenant.name == name]
        return tenant_id[0]

    def add_user_to_project(self, project_id, user_id, user_role=None):
        if not user_role:
            roles = self.keystone_client.roles.list()
            role_id = [role.id for role in roles if role.name == 'member'][0]
        if not user_id:
            user_name = cfg.common.user
            users = self.keystone_client.users.list()
            user_id = [user.id for user in users if user.name == user_name][0]
        self.keystone_client.roles.grant(role_id, user=user_id,
                                         project=project_id)

    def switch_to_project(self, name):
        projects_xpath = ("//ul[contains(@class, navbar-nav)]"
                          "//li[contains(@class, dropdown)]")
        name_xpath = ("//a//span[contains(@class, dropdown-title)"
                      "and normalize-space(text())='{0}']".format(name))
        btn_xpath = "//a[contains(@class, dropdown-toggle) and @href='#']"
        projects_list = self.driver.find_element_by_xpath(projects_xpath)
        if projects_list.text != name:
            if 'open' not in projects_list.get_attribute('class'):
                projects_list.find_element_by_xpath(btn_xpath).click()
            projects_list.find_element_by_xpath(name_xpath).click()
            self.wait_for_alert_message()
        # else the project is already set

    def take_screenshot(self, exception):
        """Taking screenshot on error

        This decorators will take a screenshot of the browser
        when the test failed or when exception raised on the test.
        Screenshot will be saved as PNG inside screenshots folder.

        """
        name = self._testMethodName
        logger.error('{0} failed'.format(name))
        screenshot_dir = './screenshots'
        if not os.path.exists(screenshot_dir):
            os.makedirs(screenshot_dir)
        filename = os.path.join(screenshot_dir, name + '.png')
        self.driver.get_screenshot_as_file(filename)

    def log_in(self, username=None, password=None):
        username = username or cfg.common.user
        password = password or cfg.common.password
        self.fill_field(by.By.ID, 'id_username', username)
        self.fill_field(by.By.ID, 'id_password', password)
        self.driver.find_element_by_xpath("//button[@type='submit']").click()
        murano = self.driver.find_element_by_xpath(consts.AppCatalog)
        if 'collapsed' in murano.get_attribute('class'):
            murano.click()

    def log_out(self):
        user_menu = self.driver.find_element(
            by.By.XPATH, "//ul[contains(@class, 'navbar-right')]")
        user_menu.find_element(
            by.By.XPATH, ".//span[@class='user-name']").click()
        user_menu.find_element(by.By.PARTIAL_LINK_TEXT, 'Sign Out').click()

    def fill_field(self, by_find, field, value):
        self.check_element_on_page(by_find, field)
        self.wait_element_is_clickable(by_find, field)
        self.driver.find_element(by=by_find, value=field).clear()
        self.driver.find_element(by=by_find, value=field).send_keys(value)

    def get_element_id(self, el_name, sec=10):
        el = ui.WebDriverWait(self.driver, sec).until(
            EC.presence_of_element_located(
                (by.By.XPATH, consts.AppPackages.format(el_name))))
        path = el.get_attribute("id")
        return path.split('__')[-1]

    def select_and_click_action_for_app(self, action, app):
        self.check_element_on_page(
            by.By.XPATH,
            "//*[@href='{0}/app-catalog/catalog/{1}/{2}']".format(
                self.url_prefix, action, app))
        self.wait_element_is_clickable(
            by.By.XPATH,
            "//*[@href='{0}/app-catalog/catalog/{1}/{2}']".format(
                self.url_prefix, action, app)).click()

    def go_to_submenu(self, link):
        element = self.wait_element_is_clickable(by.By.PARTIAL_LINK_TEXT, link)
        element.click()
        self.wait_for_sidebar_is_loaded()

    def check_panel_is_present(self, panel_name):
        self.assertIn(panel_name,
                      self.driver.find_element_by_xpath(
                          ".//*[@class='page-header']").text)

    def navigate_to(self, menu):
        el = self.wait_element_is_clickable(
            by.By.XPATH, getattr(consts, menu))
        if 'collapsed' in el.get_attribute('class'):
            el.click()

    def select_from_list(self, list_name, value, sec=10):
        locator = (by.By.XPATH,
                   "//select[contains(@name, '{0}')]"
                   "/option[@value='{1}']".format(list_name, value))
        el = ui.WebDriverWait(self.driver, sec).until(
            EC.presence_of_element_located(locator))
        el.click()

    def check_element_on_page(self, method, value, sec=10):
        try:
            ui.WebDriverWait(self.driver, sec).until(
                EC.presence_of_element_located((method, value)))
        except exc.TimeoutException:
            self.fail("Element {0} is not present on the page".format(value))

    def check_element_not_on_page(self, method, value, sec=3):
        self.driver.implicitly_wait(sec)
        present = True
        try:
            self.driver.find_element(method, value)
        except (exc.NoSuchElementException, exc.ElementNotVisibleException):
            present = False
        self.assertFalse(present, u"Element {0} is present on the page"
                                  " while it should't".format(value))
        self.driver.implicitly_wait(30)

    def check_alert_message(self, message, sec=10):
        locator = (by.By.CSS_SELECTOR, 'div.alert-dismissable')
        try:
            ui.WebDriverWait(self.driver, sec).until(
                EC.presence_of_element_located(locator))
        except exc.TimeoutException:
            self.fail("Alert is not present on the page")

        self.assertIn(message, self.driver.find_element(*locator).text)

    def create_environment(self, env_name, by_id=False):
        if by_id:
            self.driver.find_element_by_id(
                'environments__action_CreateEnvironment').click()
        else:
            self.driver.find_element_by_css_selector(
                consts.CreateEnvironment).click()
        self.fill_field(by.By.ID, 'id_name', env_name)
        self.driver.find_element_by_id(consts.ConfirmCreateEnvironment).click()
        self.wait_for_alert_message()

    def delete_environment(self, env_name, from_detail_view=False):
        if not from_detail_view:
            self.select_action_for_environment(env_name, 'delete')
        self.driver.find_element_by_xpath(consts.ConfirmDeletion).click()
        self.wait_for_alert_message()

    def select_action_for_environment(self, env_name, action):
        element_id = self.get_element_id(env_name)
        more_button = consts.More.format('environments', element_id)
        self.driver.find_element_by_xpath(more_button).click()
        btn_id = "environments__row_{0}__action_{1}".format(element_id, action)
        self.driver.find_element_by_id(btn_id).click()

    def wait_for_alert_message(self, sec=5):
        locator = (by.By.CSS_SELECTOR, 'div.alert-success')
        logger.debug("Waiting for a success message")
        ui.WebDriverWait(self.driver, sec).until(
            EC.presence_of_element_located(locator))

    def wait_for_alert_message_to_disappear(self, sec=10):
        # The alert message pops up directly over the delete environment
        # button, causing click issues. So we must wait for the alert message
        # to completely disappear before clicking the button.
        locator = (by.By.CSS_SELECTOR, 'div.alert-success')
        logger.debug("Waiting for a success message to disappear")
        ui.WebDriverWait(self.driver, sec).until(
            EC.invisibility_of_element_located(locator))

    def wait_for_error_message(self, sec=20):
        locator = (by.By.CSS_SELECTOR, 'div.alert-danger > p')
        logger.debug("Waiting for an error message")
        ui.WebDriverWait(self.driver, sec, 1).until(
            EC.presence_of_element_located(locator))
        return self.driver.find_element(*locator).text

    def wait_element_is_clickable(self, method, element, sec=10):
        return ui.WebDriverWait(self.driver, sec).until(
            EC.element_to_be_clickable((method, element)))

    def wait_for_sidebar_is_loaded(self, sec=10):
        ui.WebDriverWait(self.driver, sec).until(
            EC.presence_of_element_located(
                (by.By.CSS_SELECTOR, "nav#sidebar a.active")))
        time.sleep(0.5)

    @contextlib.contextmanager
    def wait_for_page_reload(self, sec=10):
        old_page = self.driver.find_element_by_tag_name('html')
        yield
        ui.WebDriverWait(self, sec).until(EC.staleness_of(old_page))


class PackageBase(UITestCase):
    @classmethod
    def setUpClass(cls):
        super(PackageBase, cls).setUpClass()
        cls.packages = []
        cls.mockapp_id = cls.upload_package(
            "MockApp",
            {"categories": ["Web"], "tags": ["tag"]})
        cls.postgre_id = cls.upload_package(
            "PostgreSQL",
            {"categories": ["Databases"], "tags": ["tag"]})
        cls.hot_app_id = cls.upload_package(
            "HotExample",
            {"tags": ["hot"]}, hot=True)
        cls.deployingapp_id = cls.upload_package(
            "DeployingApp",
            {"categories": ["Web"], "tags": ["tag"]},
            hot=False,
            package_dir=consts.DeployingPackageDir)

    @classmethod
    def upload_package(cls, name, data, **kwargs):
        package = utils.upload_app_package(cls.murano_client, name, data,
                                           **kwargs)
        cls.packages.append(package)
        return package

    @classmethod
    def tearDownClass(cls):
        super(PackageBase, cls).tearDownClass()
        # In case dynamically created packages are deleted at test level,
        # ignore not found errors below.
        for package in cls.packages:
            try:
                cls.murano_client.packages.delete(package)
            except muranoclient_exc.HTTPNotFound:
                pass


class ImageTestCase(PackageBase):
    @classmethod
    def setUpClass(cls):
        super(ImageTestCase, cls).setUpClass()
        glance_endpoint = cls.service_catalog.url_for(service_type='image')
        cls.glance = gclient.Client('2', endpoint=glance_endpoint,
                                    session=cls.keystone_client.session)

    def setUp(self):
        super(ImageTestCase, self).setUp()
        self.image_title = self.gen_random_resource_name('default-image', 15)
        self.image = self.upload_image(self.image_title)

    def tearDown(self):
        super(ImageTestCase, self).tearDown()
        self.glance.images.delete(self.image.id)

    @classmethod
    def upload_image(cls, title):
        try:
            murano_property = json.dumps({'title': title, 'type': 'linux'})
            image = cls.glance.images.create(name='TestImage',
                                             disk_format='qcow2',
                                             container_format='bare',
                                             is_public='True',
                                             murano_image_info=murano_property)
            image_data = BytesIO(None)
            cls.glance.images.upload(image['id'], image_data)
        except Exception:
            logger.error("Unable to create or update image in Glance")
            raise
        return image

    def select_and_click_element(self, element):
        self.driver.find_element_by_xpath(
            ".//*[@value = '{0}']".format(element)).click()


class FieldsTestCase(PackageBase):
    def check_js_error_message_is_present(self, error_message):
        self.driver.implicitly_wait(2)
        self.driver.find_element_by_xpath(
            consts.JsErrorMessage.format(error_message))
        self.driver.implicitly_wait(30)

    def check_js_error_message_is_absent(self, error_message):
        self.driver.implicitly_wait(2)
        try:
            self.driver.find_element_by_xpath(
                consts.ErrorMessage.format(error_message))
        except (exc.NoSuchElementException, exc.ElementNotVisibleException):
            logger.info("Message {0} is not"
                        " present on the page".format(error_message))

        self.driver.implicitly_wait(30)

    def check_error_message_is_present(self, error_message):
        self.driver.find_element_by_xpath(consts.ButtonSubmit).click()
        self.driver.find_element_by_xpath(
            consts.ErrorMessage.format(error_message))

    def check_error_message_is_absent(self, error_message):
        self.driver.find_element_by_xpath(consts.ButtonSubmit).click()

        self.driver.implicitly_wait(2)
        try:
            self.driver.find_element_by_xpath(
                consts.ErrorMessage.format(error_message))
        except (exc.NoSuchElementException, exc.ElementNotVisibleException):
            logger.info("Message {0} is not"
                        " present on the page".format(error_message))

        self.driver.implicitly_wait(30)


class ApplicationTestCase(ImageTestCase):
    def delete_component(self, component_name=None):
        if component_name:
            component_id = self.get_element_id(component_name)
            btn = self.wait_element_is_clickable(
                by.By.ID, 'services__row_{0}__action_delete'.format(
                    component_id))
        else:
            btn = self.wait_element_is_clickable(by.By.CSS_SELECTOR,
                                                 consts.DeleteComponent)
        btn.click()
        el = self.wait_element_is_clickable(by.By.LINK_TEXT,
                                            'Delete Component', sec=30)
        el.click()
        self.wait_for_alert_message()

    def select_action_for_package(self, package_id, action, sec=10):
        if action == 'more':
            el = self.wait_element_is_clickable(
                by.By.XPATH, "//tr[@data-object-id='{0}']"
                             "//a[@data-toggle='dropdown']".format(package_id))
            el.click()
            ui.WebDriverWait(self.driver, sec).until(lambda s: s.find_element(
                by.By.XPATH,
                ".//*[@id='packages__row_{0}__action_download_package']".
                format(package_id)).is_displayed())
        else:
            self.driver.find_element_by_xpath(
                ".//*[@id='packages__row_{0}__action_{1}']".
                format(package_id, action)).click()

    def check_package_parameter(self, selector, column, value):
        columns = {'Tenant Name': 3, 'Active': 4, 'Public': 5}
        column_num = str(columns[column])
        column_element = self.driver.find_element_by_xpath(
            "//tr[{0}]/td[{1}]".format(selector, column_num))
        self.assertTrue(column_element.text == value,
                        "'{0}' column doesn't contain '{1}'".format(column,
                                                                    value))

    def check_package_parameter_by_id(self, package_id, column, value):
        selector = '@data-object-id="{0}"'.format(package_id)
        self.check_package_parameter(selector, column, value)

    def check_package_parameter_by_name(self, package_name, column, value):
        selector = '@data-display="{0}"'.format(package_name)
        self.check_package_parameter(selector, column, value)

    def modify_package(self, param, value):
        self.fill_field(by.By.ID, 'id_{0}'.format(param), value)
        self.driver.find_element_by_xpath(consts.InputSubmit).click()
        self.wait_for_alert_message()

    def add_app_to_env(self, app_id, app_name='TestApp', env_id=None):
        self.navigate_to('Browse')
        self.go_to_submenu('Browse Local')
        if env_id:
            action = 'add'
            app = '{0}/{1}'.format(app_id, env_id)
        else:
            action = 'quick-add'
            app = app_id
        self.select_and_click_action_for_app(action, app)
        field_id = "{0}_0-name".format(app_id)
        self.fill_field(by.By.ID, field_id, value=app_name)
        self.wait_element_is_clickable(by.By.XPATH,
                                       consts.ButtonSubmit).click()
        self.wait_element_is_clickable(by.By.XPATH,
                                       consts.InputSubmit).click()
        self.select_from_list('osImage', self.image.id)

        if env_id:
            # If another app is added, then env_id is passed in. In this case,
            # the 'Next' followed by 'Create' must be clicked.
            self.check_element_on_page(by.By.CSS_SELECTOR,
                                       consts.NextWizardSubmit)
            self.wait_element_is_clickable(
                by.By.CSS_SELECTOR, consts.NextWizardSubmit).click()
            self.check_element_on_page(by.By.CSS_SELECTOR,
                                       consts.CreateWizardSubmit)
            self.wait_element_is_clickable(
                by.By.CSS_SELECTOR, consts.CreateWizardSubmit).click()

            self.wait_element_is_clickable(by.By.ID, consts.AddComponent)
            self.check_element_on_page(by.By.LINK_TEXT, app_name)
        else:
            # Otherwise, only 'Create' needs to be clicked.
            self.check_element_on_page(by.By.CSS_SELECTOR,
                                       consts.CreateWizardSubmit)
            self.wait_element_is_clickable(
                by.By.CSS_SELECTOR, consts.CreateWizardSubmit).click()
            self.wait_for_alert_message()

    def execute_action_from_table_view(self, env_name, table_action):
        """Executes an action like Deploy or Delete from the table view.

        Does not handle clicking on the confirmation modal that may appear.

        Scenario:
            1. Checks for the table drop-down button and then clicks it.
            2. Checks for the table drop-down menu to appear.
            3. Checks for the ``table_action`` button and then clicks it.
        """
        self.check_element_on_page(
            by.By.XPATH, consts.TableDropdownBtn.format(env_name))
        dropdown_btn = self.driver.find_element(
            by.By.XPATH, consts.TableDropdownBtn.format(env_name))
        dropdown_btn.click()
        self.check_element_on_page(by.By.XPATH,
                                   consts.TableDropdownMenu.format(env_name))
        self.check_element_on_page(
            by.By.XPATH, consts.TableDropdownAction.format(env_name,
                                                           table_action))
        action_btn = self.driver.find_element(
            by.By.XPATH, consts.TableDropdownAction.format(env_name,
                                                           table_action))
        action_btn.click()


class PackageTestCase(ApplicationTestCase):
    @classmethod
    def setUpClass(cls):
        super(ApplicationTestCase, cls).setUpClass()
        cls.archive_name = "ToUpload"
        cls.alt_archive_name = "ModifiedAfterUpload"
        cls.manifest = os.path.join(consts.PackageDir, 'manifest.yaml')
        cls.archive = utils.compose_package(cls.archive_name, cls.manifest,
                                            consts.PackageDir)

    def tearDown(self):
        super(PackageTestCase, self).tearDown()

        for package in self.murano_client.packages.list(include_disabled=True):
            if package.name in [self.archive_name, self.alt_archive_name]:
                self.murano_client.packages.delete(package.id)

    @classmethod
    def tearDownClass(cls):
        super(ApplicationTestCase, cls).tearDownClass()
        if os.path.exists(cls.manifest):
            os.remove(cls.manifest)
        if os.path.exists(cls.archive):
            os.remove(cls.archive)



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.core import exceptions
from unittest import mock

from muranodashboard.dynamic_ui import fields
from openstack_dashboard.test import helpers


class FlavorFlave(object):
    def __init__(self, id, name, vcpus, disk, ram):
        self.name = name
        self.vcpus = vcpus
        self.disk = disk
        self.ram = ram
        self.id = id


class TestFlavorField(helpers.APITestCase):

    @mock.patch('muranodashboard.dynamic_ui.fields.nova')
    def test_no_filter(self, mock_nova):

        # No requirements, should return all flavors
        mock_nova.novaclient().flavors.list.return_value = [
            FlavorFlave('id1', 'small', vcpus=1, disk=50, ram=1000),
            FlavorFlave('id2', 'medium', vcpus=2, disk=100, ram=2000),
            FlavorFlave('id3', 'large', vcpus=3, disk=750, ram=4000)]

        f = fields.FlavorChoiceField()
        initial_request = {}
        f.update(initial_request, self.request)
        self.assertEqual([
            ('id3', 'large'),
            ('id2', 'medium'),
            ('id1', 'small')
        ], f.choices)

    @mock.patch('muranodashboard.dynamic_ui.fields.nova')
    def test_multiple_filter(self, mock_nova):

        mock_nova.novaclient().flavors.list.return_value = [
            FlavorFlave('id2', 'medium', vcpus=2, disk=100, ram=2000),
            FlavorFlave('id3', 'large', vcpus=3, disk=750, ram=4000)]

        f = fields.FlavorChoiceField(requirements={'min_vcpus': 2})
        f.update({}, self.request)
        self.assertEqual([('id3', 'large'), ('id2', 'medium')], f.choices)

    @mock.patch('muranodashboard.dynamic_ui.fields.nova')
    def test_single_filter(self, mock_nova):

        """Check that one flavor is returned."""
        mock_nova.novaclient().flavors.list.return_value = [
            FlavorFlave('id3', 'large', vcpus=3, disk=750, ram=4000)]

        # Fake a requirement for 2 CPUs and 200 GB disk, should return large
        f = fields.FlavorChoiceField(
            requirements={'min_vcpus': 2, 'min_disk': 200})
        initial_request = {}
        f.update(initial_request, self.request)
        self.assertEqual([('id3', 'large')], f.choices)

    @mock.patch('muranodashboard.dynamic_ui.fields.nova')
    def test_no_matches_filter(self, mock_nova):

        mock_nova.novaclient().flavors.list.return_value = []

        # Fake a requirement for 4 CPUs, should return no flavors
        f = fields.FlavorChoiceField(requirements={'min_vcpus': 4})
        initial_request = {}
        f.update(initial_request, self.request)
        self.assertEqual([], f.choices)


class TestPasswordField(helpers.TestCase):
    def test_valid_input(self):
        f = fields.PasswordField('')
        for char in f.special_characters:
            if char != '\\':
                password = 'aA1111' + char
                self.assertIsNone(f.run_validators(password))

    def test_short_input(self):
        f = fields.PasswordField('')
        password = 'aA@111'
        self.assertRaises(exceptions.ValidationError, f.run_validators,
                          password)

    def test_input_without_special_characters(self):
        f = fields.PasswordField('')
        password = 'aA11111'
        self.assertRaises(exceptions.ValidationError, f.run_validators,
                          password)

    def test_input_without_digits(self):
        f = fields.PasswordField('')
        password = 'aA@@@@@'
        self.assertRaises(exceptions.ValidationError, f.run_validators,
                          password)

    def test_input_without_lowecase(self):
        f = fields.PasswordField('')
        password = 'A1@@@@@'
        self.assertRaises(exceptions.ValidationError, f.run_validators,
                          password)

    def test_input_without_uppercase(self):
        f = fields.PasswordField('')
        password = 'a1@@@@@'
        self.assertRaises(exceptions.ValidationError, f.run_validators,
                          password)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from urllib import parse

from django import forms
from django import template
from django.template import defaultfilters

register = template.Library()


@register.filter(name='is_checkbox')
def is_checkbox(field):
    return isinstance(field.field.widget, forms.CheckboxInput)


@register.filter(name='firsthalf')
def first_half(seq):
    half_len = len(seq) / 2
    return seq[:half_len]


@register.filter(name='lasthalf')
def last_half(seq):
    half_len = len(seq) / 2
    return seq[half_len:]


@register.filter(name='unquote')
@defaultfilters.stringfilter
def unquote_raw(value):
    return parse.unquote(value)



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from unittest import mock

from django.conf import settings

from muranoclient.v1 import client
from muranodashboard import api

from openstack_dashboard.test import helpers


class TestApi(helpers.APITestCase):

    def setUp(self):
        super(TestApi, self).setUp()

        factory = helpers.RequestFactoryWithMessages()
        self.request = factory.get('/path/for/testing')

        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(api, 'LOG')
    def test_handled_exceptions(self, mock_log):

        handled_exceptions = (
            (api.exc.CommunicationError,
                'Unable to communicate to murano-api server.'),
            (api.glance_exc.CommunicationError,
                'Unable to communicate to glare-api server.'),
            (api.exc.HTTPUnauthorized,
                'Check Keystone configuration of murano-api server.'),
            (api.exc.HTTPForbidden,
                'Operation is forbidden by murano-api server.'),
            (api.exc.HTTPNotFound,
                'Requested object is not found on murano server.'),
            (api.exc.HTTPConflict,
                'Requested operation conflicts with an existing object.'),
            (api.exc.BadRequest,
                'The request data is not acceptable by the server'),
            (api.exc.HTTPInternalServerError,
                'There was an error communicating with server'),
            (api.glance_exc.HTTPInternalServerError,
                'There was an error communicating with server'),
        )

        for (exception, expected_message) in handled_exceptions:
            try:
                with api.handled_exceptions(self.request):
                    raise exception()
            except exception:
                pass

            mock_log.exception.assert_called_once_with(expected_message)
            mock_log.exception.reset_mock()

    @mock.patch.object(api, 'LOG')
    def test_handled_exceptions_with_details(self, mock_log):
        exceptions_with_details = (
            (api.exc.HTTPInternalServerError,
                'There was an error communicating with server'),
            (api.glance_exc.HTTPInternalServerError,
                'There was an error communicating with server')
        )

        for (exception, expected_message) in exceptions_with_details:
            try:
                with api.handled_exceptions(self.request):
                    raise exception(details='test_details')
            except exception:
                pass

            mock_log.exception.assert_called_once_with(expected_message)
            mock_log.exception.reset_mock()

    @mock.patch.object(api, 'msg_api')
    @mock.patch.object(api, 'exceptions')
    def test_handled_exceptions_with_message_already_queued(self, mock_exc,
                                                            mock_msg_api):
        mock_message = mock.Mock(
            message='Unable to communicate to murano-api server.')
        mock_msg_api.get_messages.return_value = [mock_message]
        try:
            with api.handled_exceptions(self.request):
                raise api.exc.CommunicationError()
        except api.exc.CommunicationError:
            pass

        mock_exc.handle.assert_called_once_with(self.request, ignore=True)

    @mock.patch.object(api, 'exceptions')
    def test_handled_exceptions_with_ajax_request(self, mock_exc):
        async_messages = [('test_tag',
                           'Unable to communicate to murano-api server.',
                           'test_extra')]
        mock_request = mock.MagicMock()
        mock_request.horizon.__getitem__.return_value = async_messages
        mock_is_ajax = mock.patch(
            'horizon.utils.http.is_ajax', return_value=True).start()
        try:
            with api.handled_exceptions(mock_request):
                raise api.exc.CommunicationError()
        except api.exc.CommunicationError:
            pass

        mock_exc.handle.assert_called_once_with(mock_request, ignore=True)
        self.assertTrue(mock_is_ajax.called)
        self.assertTrue(mock_request.horizon.__getitem__.called)

    def test_muranoclient(self):
        muranoclient = api.muranoclient(self.request)
        self.assertIsNotNone(muranoclient)
        self.assertEqual(client.Client, type(muranoclient))

    @mock.patch.object(api, 'LOG')
    @mock.patch('openstack_dashboard.api.base')
    def test_muranoclient_override_endpoints(self, mock_base, mock_log):
        mock_base.url_for = mock.MagicMock(
            side_effect=api.exceptions.ServiceCatalogException)

        setattr(settings, 'MURANO_USE_GLARE', True)
        setattr(settings, 'MURANO_API_URL', None)
        setattr(settings, 'GLARE_API_URL', None)

        muranoclient = api.muranoclient(self.request)
        self.assertIsNotNone(muranoclient)
        self.assertEqual(client.Client, type(muranoclient))

        self.assertEqual(2, mock_log.warning.call_count)
        mock_log.warning.assert_any_call(
            'Murano API location could not be found in Service '
            'Catalog, using default: http://localhost:8082')
        mock_log.warning.assert_any_call(
            'Glare API location could not be found in Service '
            'Catalog, using default: http://localhost:9494')



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from muranodashboard.common import widgets

from django.core.exceptions import ValidationError
from django.core import validators
from django import forms
from django.utils.translation import gettext_lazy as _


class TriStateMultipleChoiceField(forms.ChoiceField):
    """A multiple choice checkbox field where checkboxes has three states.

    States are:

        - Checked
        - Unchecked
        - Indeterminate

    It takes a ``dict`` instance as a value,
    where keys are internal values from `choices`
    and values are ones from following (in order respectively to states):

        - True
        - False
        - None
    """
    widget = widgets.TriStateCheckboxSelectMultiple
    default_error_messages = {
        'invalid_choice': _('Select a valid choice. %(value)s is not one '
                            'of the available choices.'),
        'invalid_value': _('Enter a dict with choices and values. '
                           'Got %(value)s.'),
    }

    def to_python(self, value):
        """Checks if value, that comes from widget, is a dict."""
        if value in validators.EMPTY_VALUES:
            return {}
        elif not isinstance(value, dict):
            raise ValidationError(self.error_messages['invalid_value'],
                                  code='invalid_value')
        return value

    def validate(self, value):
        """Ensures that value has only allowed values."""
        if not set(value.keys()) <= {k for k, _ in self.choices}:
            raise ValidationError(
                self.error_messages['invalid_choice'],
                code='invalid_choice',
                params={'value': value},
            )
        elif not (set(value.values()) <=
                  set(widgets.TriStateCheckboxSelectMultiple
                      .VALUES_MAP.values())):
            raise ValidationError(
                self.error_messages['invalid_value'],
                code='invalid_value',
                params={'value': value},
            )



# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ast
from django import http as django_http
import unittest
from unittest import mock

from horizon import tables as hz_tables

from muranoclient.common import exceptions as exc
from muranodashboard.environments import consts
from muranodashboard.environments import tables
from muranodashboard.packages import consts as pkg_consts


class TestEnvironmentTables(unittest.TestCase):
    def test_check_row_actions_allowed(self):
        actions = mock.Mock()
        actions.table.data = None
        self.assertFalse(tables._check_row_actions_allowed(actions, ""))

        actions.table.data = ["test"]
        actions.allowed.return_value = False
        self.assertFalse(tables._check_row_actions_allowed(actions, ""))

        actions.allowed.return_value = True
        self.assertTrue(tables._check_row_actions_allowed(actions, ""))

    @mock.patch('muranodashboard.environments.api.deployments_list')
    def test_environment_has_deployed_services(self, deployments_list):
        deployments_list.return_value = False
        self.assertFalse(tables._environment_has_deployed_services('', ''))

        mock_deployment = mock.Mock()
        mock_deployment.description = {'services': 'service'}
        deployments_list.return_value = [mock_deployment]
        self.assertTrue(tables._environment_has_deployed_services('', ''))

        mock_deployment.description = {'services': None}
        self.assertFalse(tables._environment_has_deployed_services('', ''))

    @mock.patch('muranodashboard.environments.api.environment_get')
    def test_add_application_allowed(self, env_get):
        self.add_application = tables.AddApplication()
        self.add_application.table = mock.Mock()
        self.add_application.table.kwargs.get.return_value = "env_id"

        env_get.return_value = {'status': 'good', 'version': '1'}

        self.assertTrue(self.add_application.allowed("test", "test"))

    @mock.patch('muranodashboard.environments.tables.reverse')
    def test_add_application_get_link_url(self, reverse):
        self.add_application = tables.AddApplication()
        reverse.return_value = "reversed url"

        self.add_application.table = mock.Mock()
        self.add_application.table.kwargs = {'environment_id': 'id'}

        self.assertEqual('reversed url?next=reversed url',
                         self.add_application.get_link_url())

    def test_create_environment_allowed(self):
        self.create_environment = tables.CreateEnvironment()
        self.create_environment.table = mock.Mock()
        self.create_environment.table.data.return_value = True
        self.assertTrue(self.create_environment.allowed("test", "test"))

    @mock.patch('muranodashboard.environments.api.environment_create')
    def test_create_environment_action(self, create):
        self.create_environment = tables.CreateEnvironment()
        self.create_environment.action("", "")
        self.assertTrue(create.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.exceptions')
    def test_create_environment_action_fail(self, exceptions, reverse):
        self.create_environment = tables.CreateEnvironment()
        exceptions.handle.return_value = ""
        self.create_environment.action("", "")
        self.assertTrue(reverse.called)

    def test_delete_environment_allowed_with_environment(self):
        self.delete_environment = tables.DeleteEnvironment()
        test_environment = mock.Mock()
        test_environment.status = "test"
        self.assertTrue(self.delete_environment.allowed("", test_environment))

    def test_delete_environment_action_present(self):
        self.assertEqual('Delete Environment',
                         tables.DeleteEnvironment.action_present(1))
        self.assertEqual('Delete Environments',
                         tables.DeleteEnvironment.action_present(2))

    def test_delete_environment_action_past(self):
        self.assertEqual('Started Deleting Environment',
                         tables.DeleteEnvironment.action_past(1))
        self.assertEqual('Started Deleting Environments',
                         tables.DeleteEnvironment.action_past(2))

    @mock.patch('muranodashboard.environments.tables.'
                '_check_row_actions_allowed')
    def test_delete_environment_allowed_without_environment(self, row_actions):
        self.delete_environment = tables.DeleteEnvironment()
        row_actions.return_value = True
        self.assertTrue(self.delete_environment.allowed("", ""))

    @mock.patch('muranodashboard.environments.api.environment_delete')
    def test_delete_environment_action(self, delete):
        self.delete_environment = tables.DeleteEnvironment()
        self.delete_environment.action("", "")
        self.assertTrue(delete.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.exceptions')
    def test_delete_environment_action_fail(self, exceptions, reverse):
        self.delete_environment = tables.DeleteEnvironment()
        exceptions.handle.return_value = ""
        self.delete_environment.action("", "")
        self.assertTrue(reverse.called)

    def test_abandon_environment_is_allowed_with_environment(self):
        self.abandon_environment = tables.AbandonEnvironment()
        test_environment = mock.Mock()
        test_environment.status = "test"
        self.assertTrue(self.abandon_environment.allowed("",
                                                         test_environment))

        test_environment.status = "pending"
        self.assertFalse(self.abandon_environment.allowed("",
                                                          test_environment))

    @mock.patch('muranodashboard.environments.tables.'
                '_check_row_actions_allowed')
    def test_abandoin_environment_is_allowed_without_environment(self,
                                                                 actions):
        self.abandon_environment = tables.AbandonEnvironment()
        actions.return_value = True
        self.assertTrue(self.abandon_environment.allowed("", ""))

    @mock.patch('muranodashboard.environments.api.environment_delete')
    def test_abandon_environment_action(self, delete):
        self.abandon_environment = tables.AbandonEnvironment()
        self.abandon_environment.action("", "")
        self.assertTrue(delete.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.exceptions')
    def test_abandon_environment_action_fail(self, exceptions, reverse):
        self.abandon_environment = tables.AbandonEnvironment()
        exceptions.handle.return_value = ""
        self.abandon_environment.action("", "")
        self.assertTrue(reverse.called)

    def test_abandon_environment_action_present(self):
        self.assertEqual('Abandon Environment',
                         tables.AbandonEnvironment.action_present(1))
        self.assertEqual('Abandon Environments',
                         tables.AbandonEnvironment.action_present(2))

    def test_abandon_environment_action_past(self):
        self.assertEqual('Abandoned Environment',
                         tables.AbandonEnvironment.action_past(1))
        self.assertEqual('Abandoned Environments',
                         tables.AbandonEnvironment.action_past(2))

    @mock.patch('muranodashboard.environments.tables.'
                '_get_environment_status_and_version')
    def test_delete_service_is_allowed(self, status):
        self.delete_service = tables.DeleteService()
        status.return_value = 'test', 'test'
        self.assertTrue(self.delete_service.allowed("", ""))

    @mock.patch('muranodashboard.environments.tables.api.service_delete')
    def test_delete_service_action(self, delete):
        self.delete_service = tables.DeleteService()
        self.delete_service.table = mock.Mock()
        self.delete_service.table.kwargs.get.return_value = "test"
        self.delete_service.table.data = [{'?': {'id': 'service_id'}}]
        self.delete_service.action("", 'service_id')
        self.assertTrue(delete.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.exceptions')
    def test_delete_service_action_fail(self, exceptions, reverse):
        self.delete_service = tables.DeleteService()
        exceptions.handle.return_value = ""
        self.delete_service.action("", "")
        self.assertTrue(reverse.called)

    def test_delete_service_action_present(self):
        self.assertEqual('Delete Component',
                         tables.DeleteService.action_present(1))
        self.assertEqual('Delete Components',
                         tables.DeleteService.action_present(2))

    def test_delete_service_action_past(self):
        self.assertEqual('Started Deleting Component',
                         tables.DeleteService.action_past(1))
        self.assertEqual('Started Deleting Components',
                         tables.DeleteService.action_past(2))

    @mock.patch('muranodashboard.environments.tables.'
                '_environment_has_deployed_services')
    def test_deploy_environment_is_allowed_with_environment(self, deployed):
        self.deploy_environment = tables.DeployEnvironment()
        deployed.return_value = True
        test_environment = mock.Mock()
        test_environment.status = "pending"
        self.assertTrue(self.deploy_environment.allowed("", test_environment))

        deployed.return_value = False
        self.assertTrue(self.deploy_environment.allowed("", test_environment))

        test_environment.status = "test"
        self.assertFalse(self.deploy_environment.allowed("", test_environment))

    @mock.patch('muranodashboard.environments.tables.'
                '_check_row_actions_allowed')
    def test_deploy_environment_is_allowed_without_environment(self, actions):
        self.deploy_environment = tables.DeployEnvironment()
        actions.return_value = True
        self.assertTrue(self.deploy_environment.allowed("", ""))

    @mock.patch('muranodashboard.environments.tables.api.environment_deploy')
    def test_deploy_environment_action(self, deploy):
        self.deploy_environment = tables.DeployEnvironment()
        self.deploy_environment.action("", '')
        self.assertTrue(deploy.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.exceptions')
    def test_deploy_environment_action_fail(self, exceptions, reverse):
        self.deploy_environment = tables.DeployEnvironment()
        exceptions.handle.return_value = ""
        self.deploy_environment.action("", "")
        self.assertTrue(reverse.called)

    @mock.patch('muranodashboard.environments.tables.'
                '_get_environment_status_and_version')
    @mock.patch('muranodashboard.environments.tables.'
                '_environment_has_deployed_services')
    def test_deploy_this_environment_allowed_with_environment(self, deployed,
                                                              status):
        self.deploy_environment = tables.DeployThisEnvironment()
        deployed.return_value = True
        status.return_value = consts.STATUS_ID_READY, "version"

        self.deploy_environment.table = mock.Mock()
        self.deploy_environment.table.kwargs = {'environment_id': 'id'}

        self.assertFalse(self.deploy_environment.allowed(None, None))
        self.assertEqual('Update This Environment',
                         self.deploy_environment.verbose_name)

        deployed.return_value = False
        self.assertFalse(self.deploy_environment.allowed(None, None))
        self.assertEqual('Deploy This Environment',
                         self.deploy_environment.verbose_name)

        status.return_value = "", 0
        self.deploy_environment.table.data = None
        self.assertFalse(self.deploy_environment.allowed(None, None))

        status.return_value = "", 0
        self.deploy_environment.table.data = 'data'
        self.assertTrue(self.deploy_environment.allowed(None, None))

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.messages')
    @mock.patch('muranodashboard.environments.tables.api.environment_deploy')
    def test_deploy_this_environment_single(self, mock_deploy,
                                            mock_messages, reverse):
        self.deploy_environment = tables.DeployThisEnvironment()

        data_table = mock.Mock()
        data_table.kwargs = {'environment_id': 'id'}

        mock_deploy.side_effect = None

        self.deploy_environment.single(data_table, None, None)
        self.assertTrue(mock_messages.success.called)

    @mock.patch('muranodashboard.environments.tables.reverse')
    @mock.patch('muranodashboard.environments.tables.messages')
    @mock.patch('muranodashboard.environments.tables.api.environment_deploy')
    def test_deploy_this_environment_single_exception(self, mock_deploy,
                                                      mock_messages, reverse):
        self.deploy_environment = tables.DeployThisEnvironment()

        data_table = mock.Mock()
        data_table.kwargs = {'environment_id': 'id'}

        mock_deploy.side_effect = Exception("test")

        self.assertRaises(BaseException, self.deploy_environment.single,
                          data_table, None, None)

    def test_deploy_environment_action_present_deploy(self):
        self.assertEqual('Deploy Environment',
                         tables.DeployEnvironment.action_present_deploy(1))
        self.assertEqual('Deploy Environments',
                         tables.DeployEnvironment.action_present_deploy(2))

    def test_deploy_environment_action_past_deploy(self):
        self.assertEqual('Started deploying Environment',
                         tables.DeployEnvironment.action_past_deploy(1))
        self.assertEqual('Started deploying Environments',
                         tables.DeployEnvironment.action_past_deploy(2))

    def test_deploy_environment_action_present_update(self):
        self.assertEqual('Update Environment',
                         tables.DeployEnvironment.action_present_update(1))
        self.assertEqual('Deploy Environments',
                         tables.DeployEnvironment.action_present_update(2))

    def test_deploy_environment_action_past_update(self):
        self.assertEqual('Updated Environment',
                         tables.DeployEnvironment.action_past_update(1))
        self.assertEqual('Deployed Environments',
                         tables.DeployEnvironment.action_past_update(2))

    def test_show_environment_services(self):
        self.show_environment_services = tables.ShowEnvironmentServices()
        self.assertTrue(self.show_environment_services.allowed("", ""))

    @mock.patch.object(tables, 'reverse')
    def test_get_service_details_link(self, mock_reverse):
        mock_service = mock.MagicMock(environment_id='foo_env_id')
        mock_service.__getitem__.return_value = {'id': 'foo_service_id'}
        mock_reverse.return_value = 'test_url'

        url = tables.get_service_details_link(mock_service)
        self.assertEqual('test_url', url)
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:environments:service_details',
            args=('foo_env_id', 'foo_service_id'))

    def test_get_service_type(self):
        test_datum = {
            '?': {
                consts.DASHBOARD_ATTRS_KEY: {
                    'name': 'foo_name'
                }
            }
        }
        self.assertEqual('foo_name', tables.get_service_type(test_datum))


class TestUpdateEnvironmentRow(unittest.TestCase):
    def setUp(self):
        super(TestUpdateEnvironmentRow, self).setUp()

        self.mock_data_table = mock.Mock()
        foo_column = mock.Mock(status='foo_status')
        foo_column.configure_mock(name='foo_column')
        bar_column = mock.Mock(status='bar_status')
        bar_column.configure_mock(name='bar_column')
        self.mock_data_table.columns = {
            'foo_column': foo_column,
            'bar_column': bar_column
        }
        self.mock_data_table._meta.status_columns = [
            'foo_column', 'bar_column'
        ]
        self.mock_data_table.attrs = {}
        self.mock_datum = mock.Mock(status='foo_status')

        self.addCleanup(mock.patch.stopall)

    def test_update_environment_row(self):
        data_table = tables.UpdateEnvironmentRow(
            self.mock_data_table, self.mock_datum)
        self.assertEqual('foo_status', data_table.attrs['status'])

    @mock.patch.object(tables, 'api')
    def test_get_data(self, mock_api):
        mock_api.environment_get.side_effect = None
        mock_api.environment_get.return_value = 'test_environment'

        data_table = tables.UpdateEnvironmentRow(self.mock_data_table)
        environment = data_table.get_data(None, 'foo_environment_id')

        self.assertEqual('test_environment', environment)
        mock_api.environment_get.assert_called_once_with(
            None, 'foo_environment_id')

    @mock.patch.object(tables, 'api')
    def test_get_data_except_http_not_found(self, mock_api):
        mock_api.environment_get.side_effect = exc.HTTPNotFound

        data_table = tables.UpdateEnvironmentRow(self.mock_data_table)

        with self.assertRaises(django_http.Http404) as cm:
            data_table.get_data(None, 'foo_environment_id')
        e = cm.exception
        self.assertEqual('', str(e))

    @mock.patch.object(tables, 'api')
    def test_get_data_except_exception(self, mock_api):
        mock_api.environment_get.side_effect = Exception('foo_error')

        data_table = tables.UpdateEnvironmentRow(self.mock_data_table)

        with self.assertRaises(Exception) as cm:
            data_table.get_data(None, 'foo_environment_id')
        e = cm.exception
        self.assertEqual('foo_error', str(e))


class TestUpdateServiceRow(unittest.TestCase):

    def setUp(self):
        super(TestUpdateServiceRow, self).setUp()
        self.addCleanup(mock.patch.stopall)

    def test_update_service_row(self):
        update_service_row = tables.UpdateServiceRow(None)
        self.assertTrue(update_service_row.ajax)

    @mock.patch.object(tables, 'api')
    def test_get_data(self, mock_api):
        mock_api.service_get.return_value = 'foo_env'
        update_service_row = tables.UpdateServiceRow(mock.Mock())
        update_service_row.table.kwargs = {'environment_id': 'foo_env_id'}

        update_service_row.get_data(None, 'foo_service_id')
        mock_api.service_get.assert_called_once_with(
            None, 'foo_env_id', 'foo_service_id')


class TestUpdateEnvMetadata(unittest.TestCase):

    def test_update_env_meta_data(self):
        kwargs = {'datum': 'foo_datum'}
        update_env_meta_data = tables.UpdateEnvMetadata(**kwargs)
        self.assertEqual("update_env_metadata", update_env_meta_data.name)
        self.assertEqual("Update Metadata", update_env_meta_data.verbose_name)
        self.assertFalse(update_env_meta_data.ajax)
        self.assertEqual("pencil", update_env_meta_data.icon)
        self.assertEqual({
            "ng-controller": "MetadataModalHelperController as modal"},
            update_env_meta_data.attrs)
        self.assertTrue(update_env_meta_data.preempt)
        self.assertEqual("foo_datum", update_env_meta_data.datum)

    def test_get_link_url(self):
        update_env_meta_data = tables.UpdateEnvMetadata()
        update_env_meta_data.session_id = 'foo_session_id'
        update_env_meta_data.attrs = {}

        result = update_env_meta_data.get_link_url(mock.Mock(id='foo_env_id'))
        self.assertEqual("javascript:void(0);", result)

        lindex = update_env_meta_data.attrs['ng-click'].find('{')
        rindex = update_env_meta_data.attrs['ng-click'].rfind('}') + 1
        attrs = ast.literal_eval(update_env_meta_data.attrs['ng-click']
                                                           [lindex:rindex])
        expected_attrs = {
            'environment': 'foo_env_id',
            'session': 'foo_session_id'
        }
        update_env_meta_data.attrs['ng-click'] =\
            update_env_meta_data.attrs['ng-click'][:lindex] +\
            update_env_meta_data.attrs['ng-click'][rindex + 2:]

        self.assertEqual("modal.openMetadataModal('muranoenv', true)",
                         update_env_meta_data.attrs['ng-click'])
        for key, val in expected_attrs.items():
            self.assertEqual(val, attrs[key])

    def test_allowed(self):
        update_env_meta_data = tables.UpdateEnvMetadata()
        allowed_statuses = (
            consts.STATUS_ID_READY, consts.STATUS_ID_PENDING,
            consts.STATUS_ID_DELETE_FAILURE, consts.STATUS_ID_DEPLOY_FAILURE,
            consts.STATUS_ID_NEW)
        disallowed_statuses = (
            consts.STATUS_ID_DEPLOYING, consts.STATUS_ID_DELETING)

        for status in allowed_statuses:
            env = mock.Mock(status=status)
            self.assertTrue(update_env_meta_data.allowed(None, env))

        for status in disallowed_statuses:
            env = mock.Mock(status=status)
            self.assertFalse(update_env_meta_data.allowed(None, env))

    @mock.patch.object(tables, 'api')
    def test_update(self, mock_api):
        mock_api.Session.get_if_available.return_value = 'foo_session_id'
        update_env_meta_data = tables.UpdateEnvMetadata()
        datum = mock.Mock()
        datum.id = 'foo_env_id'
        update_env_meta_data.session_id = None

        update_env_meta_data.update(None, datum)
        self.assertEqual('foo_session_id', update_env_meta_data.session_id)
        mock_api.Session.get_if_available.assert_called_once_with(
            None, 'foo_env_id')


class TestEnvironmentsTable(unittest.TestCase):

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'policy')
    def test_environments_table(self, mock_policy, mock_reverse):
        mock_reverse.return_value = 'test_url'
        mock_env = mock.Mock(id='foo_env_id')
        mock_request = mock.Mock()
        envs_table = tables.EnvironmentsTable(mock_request)
        self.assertEqual(tables.EnvironmentsTable.get_env_detail_link.__name__,
                         envs_table.columns['name'].get_link_url.__name__)

        mock_policy.check.return_value = False
        url = envs_table.columns['name'].get_link_url(mock_env)
        self.assertIsNone(url)
        self.assertFalse(mock_reverse.called)

        mock_policy.check.return_value = True
        url = envs_table.columns['name'].get_link_url(mock_env)
        self.assertEqual('test_url', url)
        mock_reverse.assert_called_once_with(
            "horizon:app-catalog:environments:services", args=('foo_env_id',))


class TestUpdateMetadata(unittest.TestCase):

    def setUp(self):
        super(TestUpdateMetadata, self).setUp()

        update_metadata = tables.UpdateMetadata()
        self.assertEqual("update_metadata", update_metadata.name)
        self.assertEqual("Update Metadata", update_metadata.verbose_name)
        self.assertFalse(update_metadata.ajax)
        self.assertEqual("pencil", update_metadata.icon)
        self.assertIsNone(update_metadata.session_id)

    def test_get_link_url(self):
        """Test get_link_url.

        Because the dictionary part of ``attrs[ng-click]`` may have different
        key orders, extract it, convert it to an actual dict, and check that
        each value in the dict matches the expected value.
        """
        update_metadata = tables.UpdateMetadata()
        update_metadata.table = mock.Mock()
        update_metadata.table.kwargs = {'environment_id': 'foo_env_id'}
        update_metadata.session_id = 'foo_session_id'
        test_service = {
            '?': {
                'id': 'foo_service_id'
            }
        }

        result = update_metadata.get_link_url(test_service)
        self.assertEqual("javascript:void(0);", result)

        lindex = update_metadata.attrs['ng-click'].find('{')
        rindex = update_metadata.attrs['ng-click'].rfind('}') + 1
        attrs = ast.literal_eval(update_metadata.attrs['ng-click']
                                                      [lindex:rindex])
        expected_attrs = {
            'environment': 'foo_env_id',
            'session': 'foo_session_id',
            'component': 'foo_service_id'
        }
        update_metadata.attrs['ng-click'] =\
            update_metadata.attrs['ng-click'][:lindex] +\
            update_metadata.attrs['ng-click'][rindex + 2:]

        self.assertEqual("modal.openMetadataModal('muranoapp', true)",
                         update_metadata.attrs['ng-click'])
        for key, val in expected_attrs.items():
            self.assertEqual(val, attrs[key])

    @mock.patch.object(tables, 'api')
    def test_allowed(self, mock_api):
        update_metadata = tables.UpdateMetadata()
        mock_api.environment_get.return_value = mock.Mock(
            status=consts.STATUS_ID_READY)
        mock_table = mock.Mock()
        mock_table.kwargs = {'environment_id': 'foo_env_id'}
        update_metadata.table = mock_table
        self.assertTrue(update_metadata.allowed(None))

        mock_api.environment_get.return_value = mock.Mock(
            status=consts.STATUS_ID_DEPLOYING)
        self.assertFalse(update_metadata.allowed(None))
        mock_api.environment_get.assert_called_with(None, 'foo_env_id')

    @mock.patch.object(tables, 'api')
    def test_update(self, mock_api):
        mock_api.Session.get_if_available.return_value = 'foo_session_id'
        update_metadata = tables.UpdateMetadata()
        update_metadata.table = mock.Mock()
        update_metadata.table.kwargs = {'environment_id': 'foo_env_id'}
        update_metadata.session_id = None

        update_metadata.update(None, None)
        self.assertEqual('foo_session_id', update_metadata.session_id)
        mock_api.Session.get_if_available.assert_called_once_with(
            None, 'foo_env_id')


class TestServicesTable(unittest.TestCase):

    def test_get_object_id(self):
        test_datum = {'?': {'id': 'foo'}}
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        self.assertEqual('foo', services_table.get_object_id(test_datum))

    @mock.patch.object(tables, 'pkg_api')
    def test_get_apps_list(self, mock_pkg_api):
        foo_app = mock.Mock()
        foo_app.to_dict.return_value = {'foo': 'bar'}
        baz_app = mock.Mock()
        baz_app.to_dict.return_value = {'baz': 'qux'}
        mock_pkg_api.package_list.return_value = (
            [foo_app, baz_app], True
        )
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        services_table.request = None
        services_table._more = False
        expected = [{'foo': 'bar'}, {'baz': 'qux'}]

        result = services_table.get_apps_list()
        for entry in expected:
            self.assertIn(entry, result)
        mock_pkg_api.package_list.assert_called_once_with(
            None, filters={'type': 'Application', 'catalog': True})

    @mock.patch.object(tables, 'api')
    def test_actions_allowed(self, mock_api):
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        mock_api.environment_get.return_value = mock.Mock(
            status=consts.STATUS_ID_READY)
        services_table.kwargs = {'environment_id': 'foo_env_id'}
        self.assertTrue(services_table.actions_allowed())

        mock_api.environment_get.return_value = mock.Mock(
            status=consts.STATUS_ID_DEPLOYING)
        self.assertFalse(services_table.actions_allowed())
        mock_api.environment_get.assert_called_with(mock_request, 'foo_env_id')

    @mock.patch.object(tables, 'catalog_views')
    def test_categories_list(self, mock_catalog_views):
        mock_catalog_views.get_categories_list.return_value = []
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        services_table.request = None
        self.assertEqual([], services_table.get_categories_list())
        mock_catalog_views.get_categories_list.assert_called_once_with(None)

    @mock.patch('horizon.tables.actions.LinkAction.get_link_url')
    @mock.patch.object(tables, '_get_environment_status_and_version')
    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'api')
    def test_get_row_actions(
            self, mock_api, mock_reverse, mock_get_env_attrs, _):
        mock_api.extract_actions_list.return_value = [
            {'name': 'foo_bar', 'title': 'Foo Bar', 'id': 'foo_id'},
            {'name': 'baz_qux', 'title': 'Baz Qux', 'id': 'baz_id'}
        ]
        mock_reverse.return_value = 'test_url'

        mock_get_env_attrs.return_value = (consts.STATUS_ID_READY, None)
        mock_api.Session.get_if_available.return_value = 'session_id'
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        services_table.kwargs = {'environment_id': 'foo_env_id'}

        mock_datum = mock.MagicMock()
        service = {'?': {'id': 'comp_id'}}
        mock_datum.__getitem__.side_effect = lambda key: service[key]
        actions = services_table.get_row_actions(mock_datum)
        custom_actions = []

        self.assertGreater(len(actions), 0)
        for action in actions:
            if action.__class__.__name__ == 'CustomAction':
                custom_actions.append(action)
        custom_actions = sorted(custom_actions,
                                key=lambda action: action.name)
        self.assertEqual(2, len(custom_actions))

        self.assertEqual('baz_qux', custom_actions[0].name)
        self.assertEqual('Baz Qux', custom_actions[0].verbose_name)
        self.assertEqual('foo_bar', custom_actions[1].name)
        self.assertEqual('Foo Bar', custom_actions[1].verbose_name)

    @mock.patch.object(tables, '_get_environment_status_and_version')
    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'api')
    def test_get_row_actions_with_no_action_allowed_status(
            self, mock_api, mock_reverse, mock_get_env_attrs):
        mock_api.extract_actions_list.return_value = [
            {'name': 'foo_bar', 'title': 'Foo Bar', 'id': 'foo_id'},
            {'name': 'baz_qux', 'title': 'Baz Qux', 'id': 'baz_id'}
        ]
        mock_reverse.return_value = 'test_url'

        mock_get_env_attrs.return_value = (consts.STATUS_ID_DEPLOYING, None)
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        services_table.kwargs = {'environment_id': 'foo_env_id'}

        mock_datum = mock.MagicMock()
        actions = services_table.get_row_actions(mock_datum)
        self.assertEqual([], actions)

    def test_get_repo_url(self):
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        self.assertEqual(pkg_consts.DISPLAY_MURANO_REPO_URL,
                         services_table.get_repo_url())

    @mock.patch.object(tables, 'reverse')
    def test_get_pkg_def_url(self, mock_reverse):
        mock_reverse.return_value = 'test_url'
        mock_request = mock.Mock()
        services_table = tables.ServicesTable(mock_request)
        self.assertEqual('test_url',
                         services_table.get_pkg_def_url())
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')


class TestShowDeploymentDetails(unittest.TestCase):

    @mock.patch.object(tables, 'reverse')
    def test_get_link_url(self, mock_reverse):
        mock_reverse.return_value = 'test_url'
        mock_deployment = mock.Mock(
            id='foo_deployment_id', environment_id='foo_env_id')
        expected_kwargs = {'environment_id': 'foo_env_id',
                           'deployment_id': 'foo_deployment_id'}
        show_deployment_details = tables.ShowDeploymentDetails()

        url = show_deployment_details.get_link_url(mock_deployment)
        self.assertEqual('test_url', url)
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:environments:deployment_details',
            kwargs=expected_kwargs)

    def test_allowed(self):
        show_deployment_details = tables.ShowDeploymentDetails()
        self.assertTrue(show_deployment_details.allowed(None, None))


class TestEnvConfigTable(unittest.TestCase):

    def test_get_object_id(self):
        mock_request = mock.Mock()
        env_config_table = tables.EnvConfigTable(mock_request)
        self.assertEqual('foo', env_config_table.get_object_id({
            '?': {'id': 'foo'}}))


class TestDeploymentHistoryTable(unittest.TestCase):

    def setUp(self):
        super(TestDeploymentHistoryTable, self).setUp()

        deployment_history_table = tables.DeploymentHistoryTable(
            mock.Mock())
        columns = deployment_history_table.columns

        self.assertIsInstance(columns['environment_name'],
                              hz_tables.WrappingColumn)
        self.assertIsInstance(columns['logs'], hz_tables.Column)
        self.assertIsInstance(columns['services'], hz_tables.Column)
        self.assertIsInstance(columns['status'], hz_tables.Column)

        self.assertEqual('Environment', str(columns['environment_name']))
        self.assertEqual('Logs (Created, Message)', str(columns['logs']))
        self.assertEqual('Services (Name, Type)', str(columns['services']))
        self.assertEqual('Status', str(columns['status']))
        self.assertTrue(columns['status'].status)
        self.assertEqual(consts.DEPLOYMENT_STATUS_DISPLAY_CHOICES,
                         columns['status'].display_choices)

    @mock.patch.object(tables, 'template')
    def test_get_deployment_history_services(self, mock_template):
        mock_template.loader.render_to_string.return_value = \
            mock.sentinel.rendered_template

        test_description = {'services': [
            {'name': 'foo_service', '?': {'type': 'foo/bar',
                                          'name': 'foo_service'}},
            {'name': 'bar_service', '?': {'type': 'baz/qux',
                                          'name': 'bar_service'}}
        ]}
        mock_deployment = mock.Mock(description=test_description)

        result = tables.get_deployment_history_services(mock_deployment)
        self.assertEqual(mock.sentinel.rendered_template, result)

        expected_services = {
            'services': {
                'bar_service': 'baz', 'foo_service': 'foo'
            }
        }
        mock_template.loader.render_to_string.assert_called_once_with(
            'deployments/_cell_services.html', expected_services)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from muranodashboard import exceptions

ADD_INSTALLED_APPS = [
    'muranodashboard',
]

ADD_EXCEPTIONS = {
    'recoverable': exceptions.RECOVERABLE,
    'not_found': exceptions.NOT_FOUND,
    'unauthorized': exceptions.UNAUTHORIZED,
}

ADD_ANGULAR_MODULES = ['horizon.app.murano']

ADD_JS_FILES = [
    'muranodashboard/js/upload_form.js',
    'muranodashboard/js/import_bundle_form.js',
    'muranodashboard/js/more-less.js',
    'app/murano/murano.service.js',
    'app/murano/murano.module.js',
    'muranodashboard/js/add-select.js',
    'muranodashboard/js/draggable-components.js',
    'muranodashboard/js/environments-in-place.js',
    'muranodashboard/js/external-ad.js',
    'muranodashboard/js/horizon.muranotopology.js',
    'muranodashboard/js/murano.tables.js',
    'muranodashboard/js/load-modals.js',
    'muranodashboard/js/mixed-mode.js',
    'muranodashboard/js/passwordfield.js',
    'muranodashboard/js/submit-disabled.js',
    'muranodashboard/js/support_placeholder.js',
    'muranodashboard/js/validators.js'
]

FEATURE = True



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from django.views import generic

from openstack_dashboard.api.rest import urls
from openstack_dashboard.api.rest import utils as rest_utils

from muranodashboard import api
from muranodashboard.environments import api as env_api


@urls.register
class ComponentsMetadata(generic.View):
    """API for Murano components Metadata"""

    url_regex = r'app-catalog/environments/(?P<environment>[^/]+)' \
                r'/components/(?P<component>[^/]+)/metadata/$'

    @rest_utils.ajax()
    def get(self, request, environment, component):
        """Get a metadata object for a component in a given environment

        Example GET:
        http://localhost/api/app-catalog/environments/123/components/456/metadata

        The following get parameters may be passed in the GET
        request:

        :param environment: identifier of the environment
        :param component: identifier of the component

        Any additionally a "session" parameter should be passed through the API
        as a keyword.
        """
        filters, keywords = rest_utils.parse_filters_kwargs(request,
                                                            ['session'])
        session = keywords.get('session')
        if not session:
            session = env_api.Session.get_or_create_or_delete(request,
                                                              environment)
        component = api.muranoclient(request).services.get(
            environment, '/' + component, session)
        if component:
            return component.to_dict()['?'].get('metadata', {})
        return {}

    @rest_utils.ajax(data_required=True)
    def post(self, request, environment, component):
        """Set a metadata object for a component in a given environment

        Example POST:
        http://localhost/api/app-catalog/environments/123/components/456/metadata

        The following get parameters may be passed in the GET
        request:

        :param environment: identifier of the environment
        :param component: identifier of the component

        Any additionally a "session" parameter should be passed through the API
        as a keyword. Request body should contain 'updated' keyword, contain
        all the updated metadata attributes. If it is empty, the metadata is
        considered to be deleted.
        """
        client = api.muranoclient(request)
        filters, keywords = rest_utils.parse_filters_kwargs(request,
                                                            ['session'])
        session = keywords.get('session')
        if not session:
            session = env_api.Session.get_or_create_or_delete(request,
                                                              environment)
        updated = request.DATA.get('updated', {})
        path = '/{0}/%3F/metadata'.format(component)

        if updated:
            client.services.put(environment, path, updated, session)
        else:
            client.services.delete(environment, path, session)


@urls.register
class EnvironmentsMetadata(generic.View):
    """API for Murano components Metadata"""

    url_regex = r'app-catalog/environments/(?P<environment>[^/]+)/metadata/$'

    @rest_utils.ajax()
    def get(self, request, environment):
        """Get a metadata object for an environment

        Example GET:
        http://localhost/api/app-catalog/environments/123/metadata

        The following get parameters may be passed in the GET
        request:

        :param environment: identifier of the environment

        Any additionally a "session" parameter should be passed through the API
        as a keyword.
        """
        filters, keywords = rest_utils.parse_filters_kwargs(request,
                                                            ['session'])
        session = keywords.get('session')
        if not session:
            session = env_api.Session.get_or_create_or_delete(request,
                                                              environment)
        env = api.muranoclient(request).environments.get_model(
            environment, '/', session)
        if env:
            return env['?'].get('metadata', {})
        return {}

    @rest_utils.ajax(data_required=True)
    def post(self, request, environment):
        """Set a metadata object for a given environment

        Example POST:
        http://localhost/api/app-catalog/environments/123/metadata

        The following get parameters may be passed in the GET
        request:

        :param environment: identifier of the environment

        Any additionally a "session" parameter should be passed through the API
        as a keyword. Request body should contain 'updated' keyword, contain
        all the updated metadata attributes. If it is empty, the metadata is
        considered to be deleted.
        """
        client = api.muranoclient(request)
        filters, keywords = rest_utils.parse_filters_kwargs(request,
                                                            ['session'])

        session = keywords.get('session')
        if not session:
            session = env_api.Session.get_or_create_or_delete(request,
                                                              environment)
        updated = request.DATA.get('updated', {})
        patch = {
            "op": "replace",
            "path": "/?/metadata",
            "value": updated
        }
        client.environments.update_model(environment, [patch], session)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import re
import string
import types
import uuid

from django.core import validators

_LOCALIZABLE_KEYS = set(['label', 'help_text', 'error_messages'])


class ObjectID(object):
    def __init__(self):
        self.object_id = str(uuid.uuid4())


def is_localizable(keys):
    return set(keys).intersection(_LOCALIZABLE_KEYS)


def camelize(name):
    """Turns snake_case name into SnakeCase."""
    return ''.join([bit.capitalize() for bit in name.split('_')])


def decamelize(name):
    """Turns CamelCase/camelCase name into camel_case."""
    pat = re.compile(r'([A-Z]*[^A-Z]*)(.*)')
    bits = []
    while True:
        head, tail = re.match(pat, name).groups()
        bits.append(head)
        if tail:
            name = tail
        else:
            break
    return '_'.join([bit.lower() for bit in bits])


def explode(_string):
    """Explodes a string into a list of one-character strings."""
    if not _string or not isinstance(_string, str):
        return _string
    else:
        return list(_string)


def prepare_regexp(regexp):
    """Converts regular expression string pattern into RegexValidator object.

    Also /regexp/flags syntax is allowed, where flags is a string of
    one-character flags that will be appended to the compiled regexp.
    """
    if regexp.startswith('/'):
        groups = re.match(r'^/(.*)/([A-Za-z]*)$', regexp).groups()
        regexp, flags_str = groups
        flags = 0
        for flag in explode(flags_str):
            flag = flag.upper()
            if hasattr(re, flag):
                flags |= getattr(re, flag)
        return validators.RegexValidator(re.compile(regexp, flags))
    else:
        return validators.RegexValidator(re.compile(regexp))


def recursive_apply(predicate, transformer, value, *args):
    def rec(val):
        if predicate(val, *args):
            return rec(transformer(val, *args))
        elif isinstance(val, dict):
            return dict((rec(k), rec(v)) for (k, v) in val.items())
        elif isinstance(val, list):
            return [rec(v) for v in val]
        elif isinstance(val, tuple):
            return tuple([rec(v) for v in val])
        elif isinstance(val, types.GeneratorType):
            return rec(val)
        else:
            return val

    return rec(value)


def evaluate(value, context):
    return recursive_apply(
        lambda v, _ctx: hasattr(v, 'evaluate'),
        lambda v, _ctx: v.evaluate(context=_ctx),
        value, context)


def insert_hidden_ids(application):
    def wrap(k, v):
        if k == '?' and isinstance(v, dict) and not isinstance(
                v.get('id'), ObjectID):
            v['id'] = str(uuid.uuid4())
            return k, v
        return rec(k), rec(v)

    def rec(val):
        if isinstance(val, dict):
            return dict(wrap(k, v) for k, v in val.items())
        elif isinstance(val, list):
            return [rec(v) for v in val]
        elif isinstance(val, ObjectID):
            return val.object_id
        else:
            return val

    return rec(application)


def int2base(x, base):
    """Converts decimal integers to another number base from base-2 to base-36

    :param x: decimal integer
    :param base: number base, max value is 36
    :return: integer converted to the specified base
    """
    digs = string.digits + string.ascii_lowercase
    if x < 0:
        sign = -1
    elif x == 0:
        return '0'
    else:
        sign = 1
    x *= sign
    digits = []
    while x:
        digits.append(digs[x % base])
        x //= base
    if sign < 0:
        digits.append('-')
    digits.reverse()
    return ''.join(digits)


def to_str(text):
    if not isinstance(text, str):
        # unicode in python2
        if isinstance(text, str):
            text = text.encode('utf-8')
        # bytes in python3
        elif isinstance(text, bytes):
            text = text.decode('utf-8')
    return text


@contextlib.contextmanager
def current_region(request, region):
    orig_region = request.user.services_region
    if region is not None:
        request.user.services_region = region
    try:
        yield
    finally:
        request.user.services_region = orig_region



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import unittest

from muranodashboard.common import utils


class TestBunch(unittest.TestCase):
    def test_get_attr(self):
        obj = utils.Bunch(one=10)

        self.assertEqual(10, obj.one)

    def test_get_item(self):
        obj = utils.Bunch(two=15)

        self.assertEqual(15, obj['two'])

    def test_in(self):
        obj = utils.Bunch(one=10)

        self.assertIn('one', obj)

    def test_iteration(self):
        obj = utils.Bunch(one=10, two=15)

        sorted_objs = sorted([o for o in obj])

        self.assertEqual([10, 15], sorted_objs)

    def test_set_attr(self):
        obj = utils.Bunch()

        obj.one = 10

        self.assertEqual(10, obj['one'])

    def test_set_item(self):
        obj = utils.Bunch()

        obj['two'] = 20

        self.assertEqual(20, obj['two'])

    def test_del_attr(self):
        obj = utils.Bunch(one=10)

        del obj.one

        self.assertNotIn('one', obj)

    def test_del_item(self):
        obj = utils.Bunch(two=20)

        del obj['two']

        self.assertNotIn('two', obj)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ast

from django.conf import settings
from django import forms
from django.utils.translation import gettext_lazy as _
from horizon import exceptions
from horizon import forms as horizon_forms
from horizon import messages
from oslo_log import log as logging

from muranoclient.common import exceptions as exc
from muranodashboard.common import net
from muranodashboard.environments import api

LOG = logging.getLogger(__name__)
ENV_NAME_HELP_TEXT = _("Environment name must contain at least one "
                       "non-white space symbol.")


class CreateEnvironmentForm(horizon_forms.SelfHandlingForm):
    name = forms.CharField(label=_("Environment Name"),
                           help_text=ENV_NAME_HELP_TEXT,
                           max_length=255)

    net_config = horizon_forms.ThemableChoiceField(
        label=_("Environment Default Network"))

    def __init__(self, request, *args, **kwargs):
        super(CreateEnvironmentForm, self).__init__(request, *args, **kwargs)
        env_fixed_network = getattr(settings, 'USE_FIXED_NETWORK', False)
        if env_fixed_network:
            net_choices = net.get_project_assigned_network(request)
            help_text = None
            if not net_choices:
                msg = _("Default network is either not specified for "
                        "this project, or specified incorrectly, "
                        "please contact administrator.")
                messages.error(request, msg)
                raise exceptions.ConfigurationError(msg)
            else:
                self.fields['net_config'].required = False
                self.fields['net_config'].widget.attrs['readonly'] = True
        else:
            net_choices = net.get_available_networks(
                request,
                murano_networks='translate')

            if net_choices is None:  # NovaNetwork case
                net_choices = [((None, None), _('Unavailable'))]
                help_text = net.NN_HELP
            else:
                net_choices.insert(0, ((None, None), _('Create New')))
                help_text = net.NEUTRON_NET_HELP
        self.fields['net_config'].choices = net_choices
        self.fields['net_config'].help_text = help_text

    def clean(self):
        cleaned_data = super(CreateEnvironmentForm, self).clean()
        env_name = cleaned_data.get('name')
        if not env_name or not env_name.strip():
            self._errors['name'] = self.error_class([ENV_NAME_HELP_TEXT])
        return cleaned_data

    def handle(self, request, data):
        try:
            net_config = ast.literal_eval(data.pop('net_config'))
            if net_config[0] is not None:
                data.update(net.generate_join_existing_net(net_config))
            env = api.environment_create(request, data)
            request.session['env_id'] = env.id
            messages.success(request,
                             u'Created environment "{0}"'.format(data['name']))
            return True
        except exc.HTTPConflict:
            msg = _('Environment with specified name already exists')
            LOG.exception(msg)
            exceptions.handle(request, ignore=True)
            messages.error(request, msg)
            return False
        except Exception:
            msg = _('Failed to create environment')
            LOG.exception(msg)
            exceptions.handle(request)
            messages.error(request, msg)
            return False



# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
from django.conf import settings
from django import http
from django.utils.translation import gettext_lazy as _
import sys
import unittest
from unittest import mock

from horizon import conf

from muranoclient.common import exceptions as exc
from muranodashboard.environments import forms as env_forms
from muranodashboard.environments import tables as env_tables
from muranodashboard.environments import tabs as env_tabs
from muranodashboard.environments import views


@mock.patch.object(views, 'exceptions')
@mock.patch.object(views, 'api')
class TestIndexView(unittest.TestCase):

    def setUp(self):
        super(TestIndexView, self).setUp()
        self.index_view = views.IndexView()
        self.index_view.request = mock.Mock()

        self.assertEqual(env_tables.EnvironmentsTable,
                         self.index_view.table_class)
        self.assertEqual('environments/index.html',
                         self.index_view.template_name)
        self.assertEqual('Environments', self.index_view.page_title)

    def test_get_data(self, mock_api, mock_exc):
        mock_api.environments_list.return_value = ['foo_env', 'bar_env']
        environments = self.index_view.get_data()
        self.assertEqual(['foo_env', 'bar_env'], environments)

    def test_get_data_exception_communication_error(self, mock_api, mock_exc):
        mock_api.environments_list.side_effect = exc.CommunicationError
        self.index_view.get_data()
        mock_exc.handle.assert_called_once_with(
            self.index_view.request, 'Could not connect to Murano API '
                                     'Service, check connection details')

    def test_get_data_exception_http_internal_server_error(
            self, mock_api, mock_exc):
        mock_api.environments_list.side_effect = exc.HTTPInternalServerError
        self.index_view.get_data()
        mock_exc.handle.assert_called_once_with(
            self.index_view.request, 'Murano API Service is not responding. '
                                     'Try again later')

    def test_get_data_exception_http_unauthorized_error(
            self, mock_api, mock_exc):
        mock_api.environments_list.side_effect = exc.HTTPUnauthorized
        self.index_view.get_data()
        mock_exc.handle.assert_called_once_with(
            self.index_view.request, ignore=True, escalate=True)


class TestEnvironmentDetails(unittest.TestCase):

    def setUp(self):
        super(TestEnvironmentDetails, self).setUp()

        mock_request = mock.Mock()
        mock_request.user.service_catalog = None
        mock_token = mock.MagicMock()
        mock_request.user.authorized_tenants = [mock.Mock()]
        mock_token.tenant.__getitem__.return_value = 'foo_tenant_id'
        mock_request.session = {'token': mock_token}
        mock_tab_group = mock.Mock()

        self.env_details = views.EnvironmentDetails()
        self.env_details.request = mock_request
        self.env_details.tab_group_class = mock_tab_group
        self.env_details.kwargs = {'environment_id': 'foo_env_id'}

        self.assertEqual('services/index.html', self.env_details.template_name)
        self.assertEqual('{{ environment_name }}', self.env_details.page_title)

        self.addCleanup(mock.patch.stopall)

    @mock.patch.object(views, 'reverse_lazy')
    @mock.patch.object(views, 'api')
    def test_get_context_data(self, mock_api, mock_reverse_lazy):
        setattr(settings, 'MURANO_USE_GLARE', False)

        mock_env = mock.Mock()
        mock_env.configure_mock(name='foo_env')
        mock_env.id = 'foo_env_id'
        mock_deployment = mock.Mock(id='foo_deployment')
        mock_reverse_lazy.return_value = 'foo_redirect_url'

        mock_api.environment_get.return_value = mock_env
        mock_api.deployments_list.return_value = [mock_deployment]
        mock_api.deployment_reports.return_value = []

        context = self.env_details.get_context_data()
        expected_context = {
            'tab_group': self.env_details.tab_group_class(),
            'tenant_id': 'foo_tenant_id',
            'environment_name': 'foo_env',
            'poll_interval': conf.HORIZON_CONFIG['ajax_poll_interval'],
            'actions': mock.ANY,
            'url': 'foo_redirect_url',
            'view': self.env_details
        }

        for key, val in expected_context.items():
            self.assertEqual(val, context[key])
        self.assertNotIn('__action_show', context['actions'])
        self.assertNotIn('__action_deploy', context['actions'])

    @mock.patch.object(views, 'reverse_lazy')
    @mock.patch.object(views, 'exceptions')
    @mock.patch.object(views, 'api')
    def test_get_context_data_except_exception(
            self, mock_api, mock_exceptions, mock_reverse_lazy):
        mock_reverse_lazy.return_value = 'foo_redirect_url'
        mock_api.environment_get.side_effect = Exception
        expected_msg = "Sorry, this environment doesn't exist anymore"

        self.env_details.get_context_data()

        mock_exceptions.handle.assert_called_once_with(
            self.env_details.request, expected_msg,
            redirect='foo_redirect_url')

    @mock.patch.object(views, 'reverse_lazy')
    @mock.patch.object(views, 'exceptions')
    @mock.patch.object(views, 'api')
    def test_get_tabs_except_http_exception(
            self, mock_api, mock_exceptions, mock_reverse_lazy):
        mock_reverse_lazy.return_value = 'foo_redirect_url'
        mock_api.deployments_list.side_effect = exc.HTTPException
        expected_msg = "Unable to retrieve list of deployments"

        result = self.env_details.get_tabs(None)

        self.assertEqual(self.env_details.tab_group_class(), result)
        mock_exceptions.handle.assert_called_once_with(
            self.env_details.request, expected_msg,
            redirect='foo_redirect_url')
        self.env_details.tab_group_class.assert_any_call(None, logs=[])


@mock.patch.object(views, 'api')
class TestDetailServiceView(unittest.TestCase):

    def setUp(self):
        super(TestDetailServiceView, self).setUp()
        self.detail_service_view = views.DetailServiceView()
        self.detail_service_view.kwargs = {
            'service_id': 'foo_service_id',
            'environment_id': 'foo_env_id'
        }
        self.mock_request = mock.Mock(GET={})
        self.mock_request.user.service_catalog = None
        self.mock_request.horizon = {
            'async_messages': [('tag', 'msg', 'extra')]
        }
        mock.patch('horizon.utils.http.is_ajax', return_value=True).start()
        self.detail_service_view.request = self.mock_request

        self.assertEqual(env_tabs.ServicesTabs,
                         self.detail_service_view.tab_group_class)
        self.assertEqual('services/details.html',
                         self.detail_service_view.template_name)
        self.assertEqual('{{ service_name }}',
                         self.detail_service_view.page_title)

    @mock.patch('horizon.tables.views.MultiTableMixin.get_context_data')
    @mock.patch.object(views, 'reverse')
    def test_get_context_data(
            self, mock_reverse, mock_get_context_data, mock_api):
        mock_service = mock.MagicMock()
        mock_service.configure_mock(name='foo_service_name')
        mock_env = mock.Mock()
        mock_env.configure_mock(name='foo_env_name')
        mock_api.service_get.return_value = mock_service
        mock_api.environment_get.return_value = mock_env
        mock_reverse.return_value = 'foo_reverse_url'
        mock_get_context_data.return_value = {}

        context = self.detail_service_view.get_context_data()

        expected_context = {
            'service': mock_service,
            'service_name': 'foo_service_name',
            'environment_name': 'foo_env_name',
            'custom_breadcrumb': [
                ('foo_env_name', 'foo_reverse_url'),
                (_('Applications'), None)
            ]
        }

        for key, val in expected_context.items():
            self.assertEqual(val, context[key])

        self.assertEqual(mock_service, self.detail_service_view.service)
        self.assertEqual(mock_service, self.detail_service_view._service)
        mock_api.service_get.assert_any_call(
            self.mock_request, 'foo_env_id', 'foo_service_id')
        mock_api.environment_get.assert_any_call(
            self.mock_request, 'foo_env_id')

    @mock.patch.object(views, 'exceptions')
    def test_get_data_except_http_unauthorized(self, mock_exc, mock_api):
        mock_api.service_get.side_effect = exc.HTTPUnauthorized
        self.detail_service_view.get_data()
        mock_exc.handle.assert_called_once_with(self.mock_request)

    @mock.patch.object(views, 'reverse')
    @mock.patch.object(views, 'exceptions')
    def test_get_data_except_http_forbidden(self, mock_exc, mock_reverse,
                                            mock_api):
        mock_api.service_get.side_effect = exc.HTTPForbidden
        mock_reverse.return_value = 'foo_redirect_url'
        self.detail_service_view.get_data()
        mock_exc.handle.assert_called_once_with(
            self.mock_request, _('Unable to retrieve details for service'),
            redirect='foo_redirect_url')

    def test_get_tabs(self, mock_api):
        result = self.detail_service_view.get_tabs(self.mock_request)
        self.assertIsInstance(result, env_tabs.ServicesTabs)


class TestCreateEnvironmentView(unittest.TestCase):

    def setUp(self):
        super(TestCreateEnvironmentView, self).setUp()

        mock_request = mock.Mock(session={})
        mock_request.GET = {'next': 'next_foo_url'}
        mock_request.user.service_catalog = None

        self.create_env_view = views.CreateEnvironmentView()
        self.create_env_view.submit_url = 'foo_reverse_url'
        self.create_env_view.request = mock_request

        self.assertEqual(env_forms.CreateEnvironmentForm,
                         self.create_env_view.form_class)
        self.assertEqual('create_environment_form',
                         self.create_env_view.form_id)
        self.assertEqual(_('Create Environment'),
                         self.create_env_view.modal_header)
        self.assertEqual('environments/create.html',
                         self.create_env_view.template_name)
        self.assertEqual(_('Create Environment'),
                         self.create_env_view.page_title)
        self.assertEqual('environment',
                         self.create_env_view.context_object_name)
        self.assertEqual(_('Create'), self.create_env_view.submit_label)
        self.assertEqual('foo_reverse_url', self.create_env_view.submit_url)

    @mock.patch('muranodashboard.environments.forms.net')
    def test_get_form(self, mock_net):
        mock_net.get_available_networks.return_value = None
        form = self.create_env_view.get_form()

        self.assertIsInstance(form, env_forms.CreateEnvironmentForm)
        self.assertEqual('next_foo_url',
                         self.create_env_view.request.session['next_url'])

    @mock.patch.object(views, 'reverse_lazy')
    @mock.patch.object(views, 'reverse')
    def test_get_success_url(self, mock_reverse, mock_reverse_lazy):
        mock_reverse.return_value = 'foo_reverse_url'
        mock_reverse_lazy.return_value = 'foo_reverse_lazy_url'

        self.create_env_view.request.session['next_url'] = 'foo_next_url'
        self.assertEqual('foo_next_url',
                         self.create_env_view.get_success_url())

        del self.create_env_view.request.session['next_url']
        self.create_env_view.request.session['env_id'] = 'foo_env_id'
        self.assertEqual('foo_reverse_url',
                         self.create_env_view.get_success_url())
        self.assertNotIn('env_id', self.create_env_view.request.session)
        mock_reverse.assert_called_once_with(
            "horizon:app-catalog:environments:services", args=['foo_env_id'])

        self.assertEqual('foo_reverse_lazy_url',
                         self.create_env_view.get_success_url())
        mock_reverse_lazy.assert_called_once_with(
            'horizon:app-catalog:environments:index')


@mock.patch.object(views, 'reverse')
@mock.patch.object(views, 'api')
class TestDeploymentDetailsView(unittest.TestCase):

    def setUp(self):
        super(TestDeploymentDetailsView, self).setUp()

        self.mock_request = mock.Mock(session={}, GET={})
        self.mock_request.user.service_catalog = None

        self.deployment_details_view = views.DeploymentDetailsView()
        self.deployment_details_view.request = self.mock_request
        self.deployment_details_view.kwargs = {
            'deployment_id': 'foo_deployment_id',
            'environment_id': 'foo_env_id'
        }

        self.assertEqual(env_tabs.DeploymentDetailsTabs,
                         self.deployment_details_view.tab_group_class)
        self.assertEqual(env_tables.EnvConfigTable,
                         self.deployment_details_view.table_class)
        self.assertEqual('deployments/reports.html',
                         self.deployment_details_view.template_name)
        self.assertEqual('Deployment at {{ deployment_start_time }}',
                         self.deployment_details_view.page_title)

    def test_get_context_data(self, mock_api, mock_reverse):
        mock_env = mock.Mock()
        mock_env.configure_mock(name='foo_env_name')
        mock_api.muranoclient().deployments.list.return_value = []
        mock_api.environment_get.return_value = mock_env
        mock_api.get_deployment_start.return_value = 'foo_deployment_start'
        mock_reverse.return_value = 'foo_reverse_url'

        context = self.deployment_details_view.get_context_data()

        expected_context = {
            'environment_id': 'foo_env_id',
            'environment_name': 'foo_env_name',
            'deployment_start_time': 'foo_deployment_start',
            'custom_breadcrumb': [
                ('foo_env_name', 'foo_reverse_url'),
                (_('Deployments'), None)
            ]
        }

        for key, val in expected_context.items():
            self.assertEqual(val, context[key])

        mock_api.environment_get.assert_called_once_with(
            self.mock_request, 'foo_env_id')
        mock_api.get_deployment_start.assert_called_once_with(
            self.mock_request, 'foo_env_id', 'foo_deployment_id')

    def test_get_deployment(self, mock_api, mock_reverse):
        mock_api.get_deployment_descr.return_value = 'foo_deployment_descr'

        self.deployment_details_view.environment_id = 'foo_env_id'
        self.deployment_details_view.deployment_id = 'foo_deployment_id'

        self.assertEqual('foo_deployment_descr',
                         self.deployment_details_view.get_deployment())
        mock_api.get_deployment_descr.assert_called_once_with(
            self.mock_request, 'foo_env_id', 'foo_deployment_id')

    @mock.patch.object(views, 'exceptions')
    def test_get_deployment_negative(self, mock_exc, mock_api, mock_reverse):
        mock_reverse.return_value = 'foo_reverse_url'

        self.deployment_details_view.environment_id = 'foo_env_id'
        self.deployment_details_view.deployment_id = 'foo_deployment_id'

        for exception in [exc.HTTPInternalServerError, exc.HTTPNotFound]:
            mock_api.get_deployment_descr.side_effect = exception

            deployment = self.deployment_details_view.get_deployment()

            self.assertIsNone(deployment)
            mock_api.get_deployment_descr.assert_called_with(
                self.mock_request, 'foo_env_id', 'foo_deployment_id')
            mock_exc.handle.assert_called_with(
                self.mock_request,
                _("Deployment with id foo_deployment_id "
                  "doesn't exist anymore"), redirect='foo_reverse_url')

    def test_get_logs(self, mock_api, _):
        mock_api.deployment_reports.return_value = ['foo_log']

        self.deployment_details_view.environment_id = 'foo_env_id'
        self.deployment_details_view.deployment_id = 'foo_deployment_id'

        self.assertEqual(['foo_log'], self.deployment_details_view.get_logs())

    @mock.patch.object(views, 'exceptions')
    def test_get_logs_negative(self, mock_exc, mock_api, mock_reverse):
        mock_reverse.return_value = 'foo_reverse_url'

        self.deployment_details_view.environment_id = 'foo_env_id'
        self.deployment_details_view.deployment_id = 'foo_deployment_id'

        for exception in [exc.HTTPInternalServerError, exc.HTTPNotFound]:
            mock_api.deployment_reports.side_effect = exception

            logs = self.deployment_details_view.get_logs()

            self.assertEqual([], logs)
            mock_api.deployment_reports.assert_called_with(
                self.mock_request, 'foo_env_id', 'foo_deployment_id')
            mock_exc.handle.assert_called_with(
                self.mock_request,
                _("Deployment with id foo_deployment_id "
                  "doesn't exist anymore"), redirect='foo_reverse_url')

    def test_get_tabs(self, mock_api, _):
        mock_api.get_deployment_descr.return_value = 'foo_deployment_descr'
        mock_api.deployment_reports.return_value = ['foo_log']

        result = self.deployment_details_view.get_tabs(self.mock_request)
        self.assertIsInstance(result, env_tabs.DeploymentDetailsTabs)


class TestJSONView(unittest.TestCase):

    @mock.patch.object(views, 'api')
    def test_get(self, mock_api):
        mock_api.load_environment_data.return_value = "{'foo': 'bar'}"
        mock_request = mock.Mock()

        kwargs = {'environment_id': 'foo_env_id'}
        result = views.JSONView.get(mock_request, **kwargs)

        self.assertIsInstance(result, http.HttpResponse)
        self.assertEqual(b"{'foo': 'bar'}", result.content)
        mock_api.load_environment_data.assert_called_once_with(mock_request,
                                                               'foo_env_id')


class TestJSONResponse(unittest.TestCase):

    def test_init(self):
        kwargs = {'content_type': 'json'}
        json_response = views.JSONResponse(**kwargs)
        self.assertIsInstance(json_response, views.JSONResponse)
        self.assertEqual(b'{}', json_response.content)

        json_response = views.JSONResponse(content='foo', **kwargs)
        self.assertIsInstance(json_response, views.JSONResponse)
        self.assertEqual(b'"foo"', json_response.content)


class TestStartActionView(unittest.TestCase):

    @mock.patch.object(views, 'reverse')
    @mock.patch.object(views, 'api')
    def test_post(self, mock_api, mock_reverse):
        mock_api.action_allowed.return_value = True
        mock_reverse.return_value = 'foo_reverse_url'
        mock_request = mock.Mock()

        result = views.StartActionView.post(
            mock_request, 'foo_env_id', 'foo_action_id')
        self.assertIsInstance(result, views.JSONResponse)
        self.assertEqual(b'{"url": "foo_reverse_url"}', result.content)
        mock_api.run_action.assert_called_once_with(
            mock_request, 'foo_env_id', 'foo_action_id')

        mock_api.action_allowed.return_value = False
        result = views.StartActionView.post(
            mock_request, 'foo_env_id', 'foo_action_id')
        self.assertIsInstance(result, views.JSONResponse)
        self.assertEqual(b'{}', result.content)
        mock_api.action_allowed.assert_called_with(mock_request, 'foo_env_id')


class TestActionResultView(unittest.TestCase):

    def test_is_file_returned(self):
        test_result = {'result': {'?': {'type': 'io.murano.File'}}}
        self.assertTrue(views.ActionResultView.is_file_returned(test_result))

    def test_is_file_returned_negative(self):
        self.assertFalse(
            views.ActionResultView.is_file_returned({}))

    def test_compose_response(self):
        response = views.ActionResultView.compose_response('foo')
        self.assertIsInstance(response, http.HttpResponse)
        self.assertEqual(b'"foo"', response.content)
        self.assertTrue(response.has_header('Content-Disposition'))
        self.assertTrue(response.has_header('Content-Length'))
        self.assertEqual('attachment; filename=result.json',
                         response['Content-Disposition'])

    def test_compose_response_is_exc(self):
        response = views.ActionResultView.compose_response('foo', is_exc=True)
        self.assertIsInstance(response, http.HttpResponse)
        self.assertEqual(b'"foo"', response.content)
        self.assertTrue(response.has_header('Content-Disposition'))
        self.assertTrue(response.has_header('Content-Length'))
        self.assertEqual('attachment; filename=exception.json',
                         response['Content-Disposition'])

    def test_compose_response_is_file(self):
        base64_encoding = None
        if sys.version_info[0] == 2:
            base64_encoding = base64.b64encode(bytes('foo_base_64'))
        elif sys.version_info[0] == 3:
            base64_encoding = base64.b64encode(bytes('foo_base_64', 'UTF-8'))
        test_result = {
            'filename': 'filename.foo',
            'mimeType': 'foo_mime_type',
            'base64Content': base64_encoding
        }
        response = views.ActionResultView.compose_response(
            test_result, is_file=True)
        self.assertIsInstance(response, http.HttpResponse)
        self.assertEqual(b'foo_base_64', response.content)
        self.assertTrue(response.has_header('Content-Disposition'))
        self.assertTrue(response.has_header('Content-Length'))
        self.assertEqual('attachment; filename=filename.foo',
                         response['Content-Disposition'])

    @mock.patch.object(views, 'api_utils')
    def test_get(self, mock_api_utils):
        mock_api_utils.muranoclient().actions.get_result.return_value =\
            {'result': 'foo_result', 'foo': 'bar'}
        mock_request = mock.Mock()

        action_result_view = views.ActionResultView()
        result = action_result_view.get(
            mock_request, 'foo_env_id', 'foo_task_id', 'poll')
        self.assertIsInstance(result, views.JSONResponse)
        self.assertEqual(b'{"foo": "bar"}', result.content)

        mock_api_utils.muranoclient().actions.get_result.\
            assert_called_once_with('foo_env_id', 'foo_task_id')

    @mock.patch.object(views, 'api_utils')
    def test_get_with_compose_response(self, mock_api_utils):
        mock_api_utils.muranoclient().actions.get_result.return_value =\
            {'result': 'foo_result', 'isException': False}
        mock_request = mock.Mock()

        action_result_view = views.ActionResultView()
        result = action_result_view.get(
            mock_request, 'foo_env_id', 'foo_task_id', None)
        self.assertIsInstance(result, http.HttpResponse)
        self.assertEqual(b'"foo_result"', result.content)

        mock_api_utils.muranoclient().actions.get_result.\
            assert_called_once_with('foo_env_id', 'foo_task_id')

    @mock.patch.object(views, 'api_utils')
    def test_get_without_polling_result(self, mock_api_utils):
        mock_api_utils.muranoclient().actions.get_result.return_value = None
        mock_request = mock.Mock()

        action_result_view = views.ActionResultView()
        result = action_result_view.get(
            mock_request, 'foo_env_id', 'foo_task_id', None)
        self.assertIsInstance(result, views.JSONResponse)
        self.assertEqual(b'{}', result.content)

        mock_api_utils.muranoclient().actions.get_result.\
            assert_called_once_with('foo_env_id', 'foo_task_id')


class TestDeploymentHistoryView(unittest.TestCase):

    def setUp(self):
        super(TestDeploymentHistoryView, self).setUp()
        self.deployment_history_view = views.DeploymentHistoryView()

        self.mock_request = mock.Mock()
        self.deployment_history_view.request = self.mock_request
        self.deployment_history_view.environment_id = mock.sentinel.env_id

        self.assertEqual(env_tables.DeploymentHistoryTable,
                         self.deployment_history_view.table_class)
        self.assertEqual('environments/index.html',
                         self.deployment_history_view.template_name)
        self.assertEqual(_('Deployment History'),
                         self.deployment_history_view.page_title)

    @mock.patch.object(views, 'api', autospec=True)
    def test_get_data(self, mock_env_api):
        mock_env_api.deployment_history.return_value = \
            [mock.sentinel.deployment_history]

        result = self.deployment_history_view.get_data()
        self.assertEqual([mock.sentinel.deployment_history], result)

    @mock.patch.object(views, 'exceptions', autospec=True)
    @mock.patch.object(views, 'api', autospec=True)
    def test_get_data_except_http_unauthorized(self, mock_env_api,
                                               mock_exceptions):
        mock_env_api.deployment_history.side_effect = \
            exc.HTTPUnauthorized

        self.assertEqual([], self.deployment_history_view.get_data())
        mock_exceptions.handle.assert_called_once_with(self.mock_request)

    @mock.patch.object(views, 'exceptions', autospec=True)
    @mock.patch.object(views, 'reverse', autospec=True)
    @mock.patch.object(views, 'api', autospec=True)
    def test_get_data_except_http_forbidden(self, mock_env_api, mock_reverse,
                                            mock_exceptions):
        mock_env_api.deployment_history.side_effect = \
            exc.HTTPForbidden
        mock_reverse.return_value = mock.sentinel.redirect_url

        self.assertEqual([], self.deployment_history_view.get_data())
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:environments:services',
            args=[mock.sentinel.env_id])
        mock_exceptions.handle.assert_called_once_with(
            self.mock_request, _('Unable to retrieve deployment history.'),
            redirect=mock.sentinel.redirect_url)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from django.conf import settings
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy
from horizon import exceptions
from horizon import tables
from muranoclient.common import exceptions as exc
from openstack_dashboard import policy
from oslo_log import log as logging

from muranodashboard import api
from muranodashboard.common import utils as md_utils

LOG = logging.getLogger(__name__)


class AddCategory(tables.LinkAction):
    name = "add_category"
    verbose_name = _("Add Category")
    url = "horizon:app-catalog:categories:add"
    classes = ("ajax-modal",)
    icon = "plus"
    policy_rules = (("murano", "add_category"),)


class DeleteCategory(policy.PolicyTargetMixin, tables.DeleteAction):
    policy_rules = (("murano", "delete_category"),)

    @staticmethod
    def action_present(count):
        return ngettext_lazy(
            u"Delete Category",
            u"Delete Categories",
            count
        )

    @staticmethod
    def action_past(count):
        return ngettext_lazy(
            u"Deleted Category",
            u"Deleted Categories",
            count
        )

    def allowed(self, request, category=None):
        use_artifacts = getattr(settings, 'MURANO_USE_GLARE', False)
        if use_artifacts:
            return category is not None
        if category is not None:
            if not category.package_count:
                return True
        return False

    def delete(self, request, obj_id):
        try:
            api.muranoclient(request).categories.delete(obj_id)
        except exc.HTTPException:
            msg = _('Unable to delete category')
            LOG.exception(msg)
            url = reverse('horizon:app-catalog:categories:index')
            exceptions.handle(request, msg, redirect=url)


class CategoriesTable(tables.DataTable):
    name = md_utils.Column('name', verbose_name=_('Category Name'))
    use_artifacts = getattr(settings, 'MURANO_USE_GLARE', False)
    if not use_artifacts:
        package_count = tables.Column('package_count',
                                      verbose_name=_('Package Count'))

    class Meta(object):
        name = 'categories'
        verbose_name = _('Application Categories')
        table_actions = (AddCategory,)
        row_actions = (DeleteCategory,)
        multi_select = False



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools as it

from django import forms
from django.template import loader
from django.utils.encoding import force_str
from django.utils import formats

from muranodashboard.common import utils


# Widge and Input code is copied from https://github.com/gregmuellegger/
# django-floppyforms/blob/master/floppyforms/widgets.py to remove the
# dependency of django-floppyforms due to not very well supported anymore
# and was dropped from Ubuntu in Xenial.
class Widget(forms.Widget):
    is_required = False

    # Backported from Django 1.7
    @property
    def is_hidden(self):
        return self.input_type == 'hidden' \
            if hasattr(self, 'input_type') else False

    # Backported from Django 1.9
    if not hasattr(forms.Widget, 'format_value'):
        def format_value(self, value):
            return self._format_value(value)


class Input(Widget):
    template_name = 'common/tri_state_checkbox/base.html'
    input_type = None
    datalist = None

    def __init__(self, *args, **kwargs):
        datalist = kwargs.pop('datalist', None)
        if datalist is not None:
            self.datalist = datalist
        template_name = kwargs.pop('template_name', None)
        if template_name is not None:
            self.template_name = template_name
        super(Input, self).__init__(*args, **kwargs)
        # This attribute is used to inject a surrounding context in the
        # floppyforms templatetags, when rendered inside a complete form.
        self.context_instance = None

    def get_context_data(self):
        return {}

    def _format_value(self, value):
        if self.is_localized:
            value = formats.localize_input(value)
        return force_str(value)

    def get_context(self, name, value, attrs=None):
        context = {
            'type': self.input_type,
            'name': name,
            'hidden': self.is_hidden,
            'required': self.is_required,
            'True': True,
        }
        # True is injected in the context to allow stricter comparisons
        # for widget attrs. See #25.
        if self.is_hidden:
            context['hidden'] = True

        if value is None:
            value = ''

        if value != '':
            # Only add the value if it is non-empty
            context['value'] = self.format_value(value)

        context.update(self.get_context_data())
        context['attrs'] = self.build_attrs(attrs)

        for key, attr in context['attrs'].items():
            if attr == 1:
                # 1 == True so 'key="1"' will show up only as 'key'
                # Casting to a string so that it doesn't equal to True
                # See #25.
                if not isinstance(attr, bool):
                    context['attrs'][key] = str(attr)

        if self.datalist is not None:
            context['datalist'] = self.datalist
        return context

    def render(self, name, value, attrs=None, **kwargs):
        template_name = kwargs.pop('template_name', None)
        if template_name is None:
            template_name = self.template_name
        context = self.get_context(name, value, attrs=attrs or {})
        context = utils.flatten_contexts(self.context_instance, context)
        return loader.render_to_string(template_name, context)


class TriStateCheckboxSelectMultiple(Input):
    """Renders tri-state multi-selectable checkbox.

    .. note:: Subclassed from ``CheckboxSelectMultiple`` and not from
       ``SelectMultiple`` only to make
       ``horizon.templatetags.form_helpers.is_checkbox`` able to recognize
       this widget.

       Otherwise template ``horizon/common/_form_field.html`` would render
       this widget slightly incorrectly.
    """

    VALUES_MAP = {
        'True': True,
        'False': False,
        'None': None
    }

    def get_context(self, name, value, attrs=None, choices=()):
        """Renders html and JavaScript.

        :param value: Dictionary of form
                      Choice => Value (Checked|Uncheckec|Indeterminate)
        :type value: dict
        """
        context = super(TriStateCheckboxSelectMultiple, self).get_context(
            name, value, attrs
        )

        choices = dict(it.chain(self.choices, choices))
        if value is None:
            value = dict.fromkeys(choices, False)
        else:
            value = dict(dict.fromkeys(choices, False).items() +
                         value.items())

        context['values'] = [
            (choice, label, value[choice])
            for choice, label in choices.iteritems()
        ]

        return context

    @classmethod
    def parse_value(cls, value):
        """Converts encoded string with value to Python values."""
        choice, value = value.split('=')
        value = cls.VALUES_MAP[value]

        return choice, value

    def value_from_datadict(self, data, files, name):
        """Expects values in ``"key=False/True/None"`` form."""
        try:
            values = data.getlist(name)
        except AttributeError:
            if name in data:
                values = [data[name]]
            else:
                values = []

        return dict(map(self.parse_value, values))


class ExtraContextWidgetMixin(object):
    def __init__(self, *args, **kwargs):
        super(ExtraContextWidgetMixin, self).__init__(*args, **kwargs)

        self.extra_context = kwargs.pop('extra_context', {})

    def get_context(self, *args, **kwargs):
        context = super(ExtraContextWidgetMixin, self).get_context(
            *args, **kwargs
        )
        context.update(self.extra_context)
        return context



# Copyright (c) 2016 AT&T Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
from unittest import mock

from django.utils.translation import gettext_lazy as _

from muranoclient.common import exceptions as exc
from muranodashboard.packages import tables


class TestImportPackage(unittest.TestCase):

    def setUp(self):
        super(TestImportPackage, self).setUp()

        self.import_package = tables.ImportPackage()

        self.assertEqual('upload_package', self.import_package.name)
        self.assertEqual('Import Package',
                         self.import_package.verbose_name)
        self.assertEqual('horizon:app-catalog:packages:upload',
                         self.import_package.url)
        self.assertEqual(('ajax-modal',), self.import_package.classes)
        self.assertEqual('plus', self.import_package.icon)
        self.assertEqual((('murano', 'upload_package'),),
                         self.import_package.policy_rules)

    @mock.patch.object(tables, 'api')
    def test_allowed(self, mock_api):
        mock_api.muranoclient().categories.list.return_value = ['foo_cat']
        self.assertTrue(self.import_package.allowed(None, None))

        mock_api.muranoclient().categories.list.return_value = None
        self.assertFalse(self.import_package.allowed(None, None))


class TestDownloadPackage(unittest.TestCase):

    def setUp(self):
        super(TestDownloadPackage, self).setUp()

        self.download_package = tables.DownloadPackage()

        self.assertEqual('download_package', self.download_package.name)
        self.assertEqual('Download Package',
                         self.download_package.verbose_name)
        self.assertEqual((('murano', 'download_package'),),
                         self.download_package.policy_rules)
        self.assertEqual('horizon:app-catalog:packages:download',
                         self.download_package.url)

    def test_allowed(self):
        self.assertTrue(self.download_package.allowed(None, None))

    @mock.patch.object(tables, 'reverse')
    def test_get_link_url(self, mock_reverse):
        mock_reverse.return_value = 'foo_reverse_url'
        mock_app = mock.Mock(id='foo_app_id')
        mock_app.configure_mock(name='FOO APP')

        result = self.download_package.get_link_url(mock_app)

        self.assertEqual('foo_reverse_url', result)
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:download',
            args=('foo-app', 'foo_app_id'))


class TestToggleEnabled(unittest.TestCase):

    def setUp(self):
        super(TestToggleEnabled, self).setUp()

        self.mock_request = mock.Mock()
        self.toggle_enabled = tables.ToggleEnabled()
        self.toggle_enabled.request = self.mock_request

        self.assertEqual('toggle_enabled', self.toggle_enabled.name)
        self.assertEqual('Toggle Enabled', self.toggle_enabled.verbose_name)
        self.assertEqual('toggle-on', self.toggle_enabled.icon)
        self.assertEqual((('murano', 'modify_package'),),
                         self.toggle_enabled.policy_rules)

    def test_action_present(self):
        self.assertEqual('Toggle Active',
                         tables.ToggleEnabled.action_present(1))
        self.assertEqual('Toggle Active',
                         tables.ToggleEnabled.action_present(2))

    def test_action_past(self):
        self.assertEqual('Toggled Active',
                         tables.ToggleEnabled.action_past(1))
        self.assertEqual('Toggled Active',
                         tables.ToggleEnabled.action_past(2))

    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_action(self, mock_api, mock_log):
        self.toggle_enabled.action(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.toggle_active.assert_called_once_with(
            'foo_package_id')
        mock_log.debug.assert_called_once_with('Toggle Active for package '
                                               'foo_package_id.')

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'messages')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_action_except_http_forbidden(self, mock_api, mock_log, mock_exc,
                                          mock_messages, mock_reverse):
        mock_api.muranoclient().packages.toggle_active.side_effect = \
            exc.HTTPForbidden
        mock_reverse.return_value = 'foo_reverse_url'
        expected_msg = _('You are not allowed to perform this operation')

        self.toggle_enabled.action(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.toggle_active.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_msg)
        mock_messages.error.assert_called_once_with(self.mock_request,
                                                    expected_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                expected_msg,
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')


class TestTogglePublicEnabled(unittest.TestCase):

    def setUp(self):
        super(TestTogglePublicEnabled, self).setUp()

        self.mock_request = mock.Mock()
        self.toggle_public_enabled = tables.TogglePublicEnabled()
        self.toggle_public_enabled.request = self.mock_request

        self.assertEqual('toggle_public_enabled',
                         self.toggle_public_enabled.name)
        self.assertEqual('share-alt', self.toggle_public_enabled.icon)
        self.assertEqual((('murano', 'publicize_package'),),
                         self.toggle_public_enabled.policy_rules)

    def test_action_present(self):
        self.assertEqual('Toggle Public',
                         tables.TogglePublicEnabled.action_present(1))
        self.assertEqual('Toggle Public',
                         tables.TogglePublicEnabled.action_present(2))

    def test_action_past(self):
        self.assertEqual('Toggled Public',
                         tables.TogglePublicEnabled.action_past(1))
        self.assertEqual('Toggled Public',
                         tables.TogglePublicEnabled.action_past(2))

    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_action(self, mock_api, mock_log):
        self.toggle_public_enabled.action(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.toggle_public.assert_called_once_with(
            'foo_package_id')
        mock_log.debug.assert_called_once_with(
            'Toggle Public for package foo_package_id.')

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'messages')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_action_except_http_forbidden(self, mock_api, mock_log, mock_exc,
                                          mock_messages, mock_reverse):
        mock_api.muranoclient().packages.toggle_public.side_effect = \
            exc.HTTPForbidden
        mock_reverse.return_value = 'foo_reverse_url'
        expected_msg = _('You are not allowed to perform this operation')

        self.toggle_public_enabled.action(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.toggle_public.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_msg)
        mock_messages.error.assert_called_once_with(self.mock_request,
                                                    expected_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                expected_msg,
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'messages')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_action_except_http_conflict(self, mock_api, mock_log, mock_exc,
                                         mock_messages, mock_reverse):
        mock_api.muranoclient().packages.toggle_public.side_effect = \
            exc.HTTPConflict
        mock_reverse.return_value = 'foo_reverse_url'
        expected_msg = _('Package or Class with the same name is already made '
                         'public')

        self.toggle_public_enabled.action(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.toggle_public.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_msg)
        mock_messages.error.assert_called_once_with(self.mock_request,
                                                    expected_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                expected_msg,
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')


class TestDeletePackage(unittest.TestCase):

    def setUp(self):
        super(TestDeletePackage, self).setUp()

        self.mock_request = mock.Mock()
        self.delete_package = tables.DeletePackage()
        self.delete_package.request = self.mock_request

        self.assertEqual('delete_package', self.delete_package.name)
        self.assertEqual((('murano', 'delete_package'),),
                         self.delete_package.policy_rules)

    def test_action_present(self):
        self.assertEqual('Delete Package',
                         tables.DeletePackage.action_present(1))
        self.assertEqual('Delete Packages',
                         tables.DeletePackage.action_present(2))

    def test_action_past(self):
        self.assertEqual('Deleted Package',
                         tables.DeletePackage.action_past(1))
        self.assertEqual('Deleted Packages',
                         tables.DeletePackage.action_past(2))

    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_delete(self, mock_api, mock_log):

        self.delete_package.delete(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.delete.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_not_called()

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_delete_except_http_not_found(self, mock_api, mock_log, mock_exc,
                                          mock_reverse):
        mock_api.muranoclient().packages.delete.side_effect = exc.HTTPNotFound
        mock_reverse.return_value = 'foo_reverse_url'
        expected_msg = _('Package with id foo_package_id is not found')

        self.delete_package.delete(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.delete.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                expected_msg,
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_delete_except_http_forbidden(self, mock_api, mock_log, mock_exc,
                                          mock_reverse):
        mock_api.muranoclient().packages.delete.side_effect = exc.HTTPForbidden
        mock_reverse.return_value = 'foo_reverse_url'
        expected_msg = _('You are not allowed to delete this package')

        self.delete_package.delete(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.delete.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                expected_msg,
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')

    @mock.patch.object(tables, 'reverse')
    @mock.patch.object(tables, 'exceptions')
    @mock.patch.object(tables, 'LOG')
    @mock.patch.object(tables, 'api')
    def test_delete_except_exception(self, mock_api, mock_log, mock_exc,
                                     mock_reverse):
        mock_api.muranoclient().packages.delete.side_effect = Exception
        mock_reverse.return_value = 'foo_reverse_url'
        expected_log_msg = _('Unable to delete package in murano-api server')

        self.delete_package.delete(self.mock_request, 'foo_package_id')

        mock_api.muranoclient().packages.delete.assert_called_once_with(
            'foo_package_id')
        mock_log.exception.assert_called_once_with(expected_log_msg)
        mock_exc.handle.assert_called_once_with(self.mock_request,
                                                _('Unable to remove package.'),
                                                redirect='foo_reverse_url')
        mock_reverse.assert_called_once_with(
            'horizon:app-catalog:packages:index')


class TestModifyPackage(unittest.TestCase):

    def setUp(self):
        super(TestModifyPackage, self).setUp()

        self.modify_package = tables.ModifyPackage()

        self.assertEqual('modify_package', self.modify_package.name)
        self.assertEqual(_('Modify Package'),
                         self.modify_package.verbose_name)
        self.assertEqual('horizon:app-catalog:packages:modify',
                         self.modify_package.url)
        self.assertEqual(('ajax-modal',), self.modify_package.classes)
        self.assertEqual('edit', self.modify_package.icon)
        self.assertEqual((('murano', 'modify_package'),),
                         self.modify_package.policy_rules)

    def test_allowed(self):
        self.assertTrue(self.modify_package.allowed(None, None))
